from abc import ABC, abstractmethod
from contextlib import contextmanager


class AbstractOutputWriter(ABC):

    def __init__(self):
        self._prefix = ''
        self._ignore_prefix = False

    @abstractmethod
    def _write(self, msg: str, newline: bool = True):
        raise NotImplementedError

    def write(self, msg: str = '', newline: bool = True):
        if not self._ignore_prefix:
            msg = '\n'.join([self._prefix + line for line in msg.split('\n')])

        self._write(msg, newline=newline)
        self._ignore_prefix = not newline

    @contextmanager
    def prefix(self, prefix):
        self._prefix += prefix
        yield
        self._prefix = self._prefix[:-len(prefix)]



from typing import List
from spotml.config.abstract_instance_config import AbstractInstanceConfig, VolumeMount
from spotml.config.abstract_instance_volume import AbstractInstanceVolume
from spotml.providers.gcp.config.disk_volume import DiskVolume
from spotml.providers.gcp.config.validation import validate_instance_parameters


VOLUME_TYPE_DISK = 'Disk'
DEFAULT_IMAGE_NAME = 'spotml'


class InstanceConfig(AbstractInstanceConfig):

    def _validate_instance_params(self, params: dict) -> dict:
        return validate_instance_parameters(params)

    def _get_instance_volumes(self) -> List[AbstractInstanceVolume]:
        volumes = []
        for volume_config in self._params['volumes']:
            volume_type = volume_config['type']
            if volume_type == DiskVolume.TYPE_NAME:
                volumes.append(DiskVolume(volume_config, self.project_config.project_name, self.name))
            else:
                raise ValueError('GCP volume type "%s" not supported.' % volume_type)

        return volumes

    @property
    def user(self):
        return 'spotml'

    @property
    def machine_name(self) -> str:
        """Name of the Compute Engine instance."""
        return '%s-%s' % (self.project_config.project_name.lower(), self.name.lower())

    @property
    def project_id(self) -> str:
        return self._params['projectId']

    @property
    def zone(self) -> str:
        return self._params['zone']

    @property
    def machine_type(self) -> str:
        return self._params['machineType']

    @property
    def gpu(self) -> dict:
        return self._params['gpu']

    @property
    def is_preemptible_instance(self) -> bool:
        return self._params['preemptibleInstance']

    @property
    def boot_disk_size(self) -> int:
        return self._params['bootDiskSize']

    @property
    def ports(self) -> List[int]:
        return list(set(self._params['ports']))

    @property
    def image_name(self) -> str:
        return self._params['imageName']

    @property
    def has_image_name(self) -> bool:
        return bool(self._params['imageName'])

    @property
    def image_uri(self) -> str:
        return self._params['imageUri']



from typing import List
from spotml.config.validation import is_subdir


PROJECT_VOLUME_MOUNT_NAME = '.project'


class ContainerConfig(object):

    def __init__(self, container_config: dict):
        self._config = container_config
        self._volume_mounts = self._get_volume_mounts()

    @property
    def name(self) -> str:
        return self._config['name']

    @property
    def project_dir(self) -> str:
        return self._config['projectDir']

    @property
    def image(self) -> str:
        return self._config['image']

    @property
    def file(self) -> str:
        return self._config['file']

    @property
    def run_as_host_user(self) -> str:
        return self._config['runAsHostUser']

    @property
    def volume_mounts(self) -> list:
        return self._volume_mounts

    @property
    def commands(self) -> str:
        return self._config['commands']

    @property
    def working_dir(self) -> str:
        """Working directory for the Docker container."""
        working_dir = self._config['workingDir']
        if not working_dir:
            working_dir = self._config['projectDir']

        return working_dir

    @property
    def env(self) -> dict:
        return self._config['env']

    @property
    def host_network(self) -> bool:
        return self._config['hostNetwork']

    @property
    def ports(self) -> List[dict]:
        return self._config['ports']

    @property
    def runtime_parameters(self) -> list:
        return self._config['runtimeParameters']

    def _get_volume_mounts(self):
        """Returns container volume mounts from the configuration and
        adds the project volume mount if necessary."""

        volume_mounts = self._config['volumeMounts']

        # check if the project directory is a sub-directory of one of the volume mounts
        project_has_volume = False
        for volume_mount in volume_mounts:
            if is_subdir(self.project_dir, volume_mount['mountPath']):
                project_has_volume = True
                break

        # if it's not, then add new volume mount
        if not project_has_volume:
            volume_mounts.insert(0, {
                'name': PROJECT_VOLUME_MOUNT_NAME,
                'mountPath': self.project_dir,
            })

        return volume_mounts



from spotml.commands.writers.abstract_output_writrer import AbstractOutputWriter
from spotml.errors.nothing_to_do import NothingToDoError
from spotml.deployment.abstract_docker_instance_manager import AbstractDockerInstanceManager
from spotml.providers.local.config.instance_config import InstanceConfig


class InstanceManager(AbstractDockerInstanceManager):

    instance_config: InstanceConfig

    def _get_instance_config(self, instance_config: dict) -> InstanceConfig:
        """Validates the instance config and returns an InstanceConfig object."""
        return InstanceConfig(instance_config, self.project_config)

    def is_running(self):
        return True

    def clean(self, output: AbstractOutputWriter):
        pass

    def sync(self, output: AbstractOutputWriter, dry_run=False):
        raise NothingToDoError('Nothing to do. The project directory is mounted to the container.')

    def download(self, download_filters: list, output: AbstractOutputWriter, dry_run=False):
        raise NothingToDoError('Nothing to do. The project directory is mounted to the container.')



import shlex
from spotml.deployment.container.abstract_container_commands import AbstractContainerCommands
from spotml.deployment.utils.cli import shlex_join


class DockerCommands(AbstractContainerCommands):

    def build(self, image_name: str) -> str:
        if not self._instance_config.dockerfile_path:
            raise ValueError('Cannot generate the "build" command as Dockerfile path is not specified')

        if not self._instance_config.docker_context_path:
            raise ValueError('Cannot generate the "build" command as Docker context path is not set')

        build_cmd = 'docker build -t %s -f %s %s' % (image_name, shlex.quote(self._instance_config.dockerfile_path),
                                                     shlex.quote(self._instance_config.docker_context_path))

        if self._instance_config.container_config.run_as_host_user:
            build_cmd += ' --build-arg USER_ID=$(id -u %s) --build-arg GROUP_ID=$(id -g %s)' \
                         % (self._instance_config.user, self._instance_config.user)

        return build_cmd

    def pull(self) -> str:
        return 'docker pull ' + self._instance_config.container_config.image

    def run(self, image_name: str = None) -> str:
        image_name = image_name if image_name else self._instance_config.container_config.image

        # prepare "docker run" arguments
        args = ['-td'] + self._instance_config.container_config.runtime_parameters

        if self._instance_config.container_config.host_network:
            args += ['--net=host']

        for port in self._instance_config.container_config.ports:
            host_port = port['hostPort']
            container_port = port['containerPort']
            args += ['-p', ('%d:%d' % (host_port, container_port)) if host_port else str(container_port)]

        for volume_mount in self._instance_config.volume_mounts:
            args += ['-v', '%s:%s:%s' % (volume_mount.host_path, volume_mount.mount_path, volume_mount.mode)]

        for env_name, env_value in self._instance_config.container_config.env.items():
            args += ['-e', '%s=%s' % (env_name, env_value)]

        args += ['--name', self._instance_config.full_container_name]

        run_cmd = 'docker run $(nvidia-smi &> /dev/null && echo "--gpus all")'

        if self._instance_config.container_config.run_as_host_user:
            run_cmd += ' -u $(id -u %s):$(id -g %s) -e HOST_USER_ID=$(id -u %s) -e HOST_GROUP_ID=$(id -g %s)' \
                       % tuple([self._instance_config.user] * 4)

        run_cmd += ' %s %s /bin/sh > /dev/null' % (shlex_join(args), image_name)

        return run_cmd

    def is_created(self, container_name: str = None, is_running: bool = False) -> str:
        container_name = container_name if container_name else self._instance_config.full_container_name
        show_all = '' if is_running else 'a'

        test_cmd = '[ $(docker ps -q%s --filter name="%s" | wc -c) -ne 0 ]' % (show_all, container_name)

        return test_cmd

    def remove(self):
        return 'docker rm -f "%s" > /dev/null' % self._instance_config.full_container_name

    def exec(self, command: str, interactive: bool = False, tty: bool = False, user: str = None,
             container_name: str = None, working_dir: str = None) -> str:
        container_name = container_name if container_name else self._instance_config.full_container_name
        working_dir = working_dir if working_dir else self._instance_config.container_config.working_dir

        exec_cmd = 'docker exec'

        if interactive:
            exec_cmd += ' -i'

        if tty:
            exec_cmd += ' -t'

        if user:
            exec_cmd += ' -u ' + shlex.quote(user)

        if working_dir:
            # no quoting, it can be environmental variable
            exec_cmd += ' -w ' + working_dir

        exec_cmd += ' %s %s' % (container_name, command)

        # run "exec" command only if the container is running
        test_cmd = self.is_created(container_name, is_running=True)
        error_msg = 'Container is not running.\\nUse the "spotml start -C" command to start it.\\n'
        cond_exec_cmd = 'if %s; then %s; else printf %s; exit 1; fi' % (test_cmd, exec_cmd, shlex.quote(error_msg))

        return cond_exec_cmd



from spotml.deployment.abstract_cloud_instance.resources.abstract_bucket import AbstractBucket
from google.cloud.storage import Bucket as GSBucket


class Bucket(AbstractBucket):

    def __init__(self, bucket: GSBucket):
        self._bucket = bucket

    @property
    def name(self) -> str:
        return self._bucket.name



import boto3
import requests as requests

from spotml.config.config_utils import DEFAULT_CONFIG_FILENAME
from spotml.config.url import API_URL
from spotml.constants.constants import GENERIC_ERROR_MESSAGE
from spotml.deployment.abstract_cloud_instance.abstract_cloud_instance_manager import AbstractCloudInstanceManager
from spotml.errors.instance_not_running import InstanceNotRunningError
from spotml.providers.aws.config.instance_config import InstanceConfig
from spotml.providers.aws.data_transfer import DataTransfer
from spotml.providers.aws.instance_deployment import InstanceDeployment
from spotml.providers.aws.resource_managers.bucket_manager import BucketManager
from spotml.services.run_service import get_instance_tracking_status
from spotml.utils import render_table
from tabulate import tabulate
from dateutil import parser
from tzlocal import get_localzone


def get_local_time_from_string(utc_date_time_string):
    return parser.parse(utc_date_time_string).astimezone(get_localzone()).strftime('%b %d, %Y -%l:%M%p %Z')


class InstanceManager(AbstractCloudInstanceManager):
    instance_config: InstanceConfig
    bucket_manager: BucketManager
    data_transfer: DataTransfer
    instance_deployment: InstanceDeployment

    def _get_instance_config(self, instance_config: dict) -> InstanceConfig:
        """Validates the instance config and returns an InstanceConfig object."""
        return InstanceConfig(instance_config, self.project_config)

    def _get_bucket_manager(self) -> BucketManager:
        """Returns an bucket manager."""
        return BucketManager(self.instance_config.project_config.project_name, self.instance_config.region)

    def _get_data_transfer(self) -> DataTransfer:
        """Returns a data transfer object."""
        return DataTransfer(
            local_project_dir=self.project_config.project_dir,
            host_project_dir=self.instance_config.host_project_dir,
            sync_filters=self.project_config.sync_filters,
            instance_name=self.instance_config.name,
            region=self.instance_config.region,
        )

    def _get_instance_deployment(self) -> InstanceDeployment:
        """Returns an instance deployment manager."""
        return InstanceDeployment(self.instance_config)

    def sendConfigCredentialFiles(self, script_name: str):
        session = boto3.session.Session()
        credentials = session.get_credentials()

        region = session.region_name
        if not region:
            raise ValueError("You must specify a default region. You can configure your region by "
                             "running 'aws configure'")

        access_key = credentials.access_key
        secret_key = credentials.secret_key

        with open(DEFAULT_CONFIG_FILENAME, 'r') as spotYml:
            with open(self.ssh_key_path, 'r') as sshKey:
                files = {'spotml_config': spotYml, 'ssh_key': sshKey}
                r = requests.post(f'{API_URL}/run', files=files,
                                  data={'script_name': script_name, 'region': region,
                                        'access_key': access_key, 'secret_key': secret_key})
                if r.status_code != 201:
                    print("Oops, looks like something went wrong in scheduling the run")
                    print(GENERIC_ERROR_MESSAGE)
                    raise Exception('Something went wrong in scheduling run in cloud', r)
                old_runs = r.json()['old_runs']
                if len(old_runs) > 0:
                    print("There is already an old scheduled run. Below run will be stopped and replaced by the "
                          "current run.")
                    print("OLD RUN")
                    table = []
                    for run in old_runs:
                        table.append([run['id'], run['script_name'], run['status'],
                                      get_local_time_from_string(run['created_date'])])
                    print(tabulate(table, headers=['Run Id', 'Command', 'Status', 'Created Date'], tablefmt="psql"))

                print("\n CURRENT RUN")
                current_run = r.json()['current_run']
                print(tabulate(
                    [[current_run['id'], current_run['script_name'], current_run['status'],
                      get_local_time_from_string(current_run['created_date'])]],
                    headers=['Run Id', 'Command', 'Status', 'Created Date'], tablefmt="psql"))

    def get_status_text(self) -> str:
        instance = self.instance_deployment.get_instance()
        if not instance:
            table = [
                ('Name', 'Status'),
                (self.instance_config.name, 'Not Running'),
            ]
            return render_table(table)

        table = [
            ('Instance State', instance.state),
            ('Instance Type', instance.instance_type),
            ('Availability Zone', instance.availability_zone),
        ]

        if instance.public_ip_address:
            table.append(('Public IP Address', instance.public_ip_address))

        if instance.lifecycle == 'spot':
            spot_price = instance.get_spot_price()
            table.append(('Purchasing Option', 'Spot Instance'))
            table.append(('Spot Instance Price', '$%.04f' % spot_price))
        else:
            on_demand_price = instance.get_on_demand_price()
            table.append(('Purchasing Option', 'On-Demand Instance'))
            table.append(('Instance Price', ('$%.04f (us-east-1)' % on_demand_price) if on_demand_price else 'Unknown'))

        is_tracking_idle_time = get_instance_tracking_status()
        table.append(('Tracking Idle Time', is_tracking_idle_time))

        return render_table(table)

    @property
    def ssh_key_path(self):
        return self.instance_deployment.key_pair_manager.key_path



from abc import abstractmethod
from typing import List
from argparse import Namespace, ArgumentParser
from spotml.config.config_utils import load_config
from spotml.deployment.abstract_instance_manager import AbstractInstanceManager
from spotml.providers.instance_manager_factory import InstanceManagerFactory
from spotml.commands.abstract_command import AbstractCommand
from spotml.commands.writers.abstract_output_writrer import AbstractOutputWriter


class AbstractConfigCommand(AbstractCommand):
    """Abstract class for a Spotml sub-command that needs to use a project's configuration."""

    @abstractmethod
    def _run(self, instance_manager: AbstractInstanceManager, args: Namespace, output: AbstractOutputWriter):
        raise NotImplementedError

    def configure(self, parser: ArgumentParser):
        super().configure(parser)
        parser.add_argument('-c', '--config', type=str, default=None, help='Path to the configuration file')
        parser.add_argument('instance_name', metavar='INSTANCE_NAME', nargs='?', type=str, help='Instance name')

    def run(self, args: Namespace, output: AbstractOutputWriter):
        # get project configuration
        project_config = load_config(args.config)

        # get instance configuration
        instance_id = self._get_instance_id(project_config.instances, args.instance_name, output)
        instance_config = project_config.instances[instance_id]

        # create an instance manger
        instance_manager = InstanceManagerFactory.get_instance(project_config, instance_config)

        # run the command
        self._run(instance_manager, args, output)

    @staticmethod
    def _get_instance_id(instances: List[dict], instance_name: str, output: AbstractOutputWriter):
        if not instance_name:
            if len(instances) > 1:
                # ask user to choose the instance
                output.write('Select the instance:\n')
                with output.prefix('  '):
                    for i, instance_config in enumerate(instances):
                        output.write('[%d] %s' % (i + 1, instance_config['name']))
                output.write()

                try:
                    num = int(input('Enter number: '))
                    output.write()
                except ValueError:
                    num = 0

                if num < 1 or num > len(instances):
                    raise ValueError('The value from 1 to %d was expected.' % len(instances))

                instance_id = num - 1
            else:
                instance_id = 0
        else:
            # get instance ID by name
            instance_ids = [i for i, instance in enumerate(instances) if instance['name'] == instance_name]
            if not instance_ids:
                raise ValueError('Instance "%s" not found in the configuration file' % instance_name)

            instance_id = instance_ids[0]

        return instance_id



from spotml.providers.gcp.config.image_uri import ImageUri
from spotml.providers.gcp.helpers.ce_client import CEClient


class Image(object):

    def __init__(self, data: dict):
        """
        Args:
            data (dict): Example:
                {'archiveSizeBytes': '3652446976',
                 'creationTimestamp': '2018-11-03T19:00:48.577-07:00',
                 'description': '',
                 'diskSizeGb': '10',
                 'guestOsFeatures': [{'type': 'VIRTIO_SCSI_MULTIQUEUE'}],
                 'id': '7541350343606791231',
                 'kind': 'compute#image',
                 'labelFingerprint': '42WmSpB8rSM=',
                 'licenseCodes': ['1000201'],
                 'licenses': ['https://www.googleapis.com/compute/v1/projects/ubuntu-os-cloud/global/licenses/ubuntu-1604-xenial'],
                 'name': 'spotml-ami',
                 'selfLink': 'https://www.googleapis.com/compute/v1/projects/spotml-221422/global/images/spotml-ami',
                 'sourceDisk': 'https://www.googleapis.com/compute/v1/projects/spotml-221422/zones/us-central1-a/disks/spotml-ami',
                 'sourceDiskId': '3401548142858207031',
                 'sourceType': 'RAW',
                 'status': 'READY'}
        """
        self._data = data

    @staticmethod
    def get_by_name(ce: CEClient, image_name: str):
        """Returns an image by its name."""
        res = ce.list_images(image_name)
        if not res:
            return None

        return Image(res[0])

    @staticmethod
    def get_by_uri(ce: CEClient, image_uri: str):
        image_uri = ImageUri(image_uri)
        if image_uri.is_family:
            image_data = ce.get_image_from_family(family_name=image_uri.name, project_id=image_uri.project_id)
        else:
            res = ce.list_images(image_name=image_uri.name, project_id=image_uri.project_id)
            image_data = res[0] if res else None

        if not image_data:
            return None

        return Image(image_data)

    @property
    def image_id(self) -> str:
        return self._data['id']

    @property
    def name(self) -> str:
        return self._data['name']

    @property
    def size(self) -> int:
        return self._data['diskSizeGb']

    @property
    def self_link(self) -> str:
        return self._data['selfLink']

    @property
    def source_disk(self):
        return self._data['sourceDisk']



from argparse import ArgumentParser, Namespace

from spotml.commands.abstract_config_command import AbstractConfigCommand
from spotml.commands.writers.abstract_output_writrer import AbstractOutputWriter
from spotml.constants.constants import INTERNAL_TAG
from spotml.constants.run_status import RUNNING, LAUNCHING_INSTANCE, STOPPED_BY_USER, SCHEDULED, WAITING_FOR_SPOT
from spotml.deployment.abstract_instance_manager import AbstractInstanceManager
from spotml.deployment.utils.commands import get_tmux_session_command, \
    get_bash_command, get_send_keys_command, get_script_command
from spotml.deployment.utils.user_scripts import parse_script_parameters, render_script
from spotml.errors.nothing_to_do import NothingToDoError
from spotml.services.run_service import update_run_status, get_run


class HandleScheduledRunCommand(AbstractConfigCommand):
    name = 'handle-scheduled-run'
    description = 'Run a custom script from the configuration file inside the container'

    def configure(self, parser: ArgumentParser):
        super().configure(parser)
        parser.add_argument('script_name', metavar='SCRIPT_NAME', type=str, help='Script name')
        parser.add_argument('-u', '--user', type=str, default=None,
                            help='Container username or UID (format: <name|uid>[:<group|gid>])')
        parser.add_argument('-s', '--session-name', type=str, default=None, help='tmux session name')
        parser.add_argument('-l', '--logging', action='store_true', help='Log the script outputs to a file')
        parser.add_argument('-p', '--parameter', metavar='PARAMETER=VALUE', action='append', type=str, default=[],
                            help='Set a value for the script parameter (format: PARAMETER=VALUE). This '
                                 'argument can be used multiple times to set several parameters. Parameters can be '
                                 'used in the script as Mustache variables (for example: {{PARAMETER}}).')
        parser.add_argument('--no-sync', action='store_true', help='Don\'t sync the project before running the script')

        # add the "double-dash" argument to the usage message
        parser.prog = 'spotml run'
        parser.usage = parser.format_usage()[7:-1] + ' [-- args...]\n'
        parser.epilog = 'The double dash (--) separates custom arguments that you can pass to the script ' \
                        'from the spotml arguments.'

    def _run(self, instance_manager: AbstractInstanceManager, args: Namespace, output: AbstractOutputWriter):
        run_status = get_run(args.run_id)['status']
        # Either it's a new scheduled run OR it's waiting for spot OR user reran a new run on a running instance.
        if run_status != SCHEDULED and run_status != WAITING_FOR_SPOT and run_status != RUNNING:
            print(f'{INTERNAL_TAG}Run status: {run_status} is neither "Scheduled", "Waiting for spot" nor "Running". '
                  f'Exiting now')
            return
        script_name = args.script_name
        scripts = instance_manager.project_config.scripts
        if script_name not in scripts:
            raise ValueError('Script "%s" is not defined in the configuration file.' % script_name)

        # replace script parameters
        params = parse_script_parameters(args.parameter)
        script_content = render_script(scripts[script_name], params)

        # check that the instance is started
        if not instance_manager.is_running():
            update_run_status(args.run_id, LAUNCHING_INSTANCE)
            instance_manager.start(output)

            instance_name = ''
            if len(instance_manager.project_config.instances) > 1:
                instance_name = ' ' + instance_manager.instance_config.name

            output.write('\n%s\n'
                         '\nUse the "spotml sh%s" command to connect to the container.\n'
                         % (instance_manager.get_status_text(), instance_name))

        # sync the project with the instance
        if not args.no_sync:
            try:
                instance_manager.sync_to_instance(output)
            except NothingToDoError:
                pass

        session_name = 'run'
        # get a command to run the script with "docker exec"
        # send_keys_command = get_send_keys_command(script_content, session_name)
        script_command = get_script_command(script_name, script_content, script_args=args.custom_args,
                                            logging=args.logging)
        command = instance_manager.container_commands.exec(script_command, interactive=True, tty=True,
                                                           user=args.user)

        # wrap the command with the tmux session
        if instance_manager.use_tmux:
            default_command = instance_manager.container_commands.exec(get_bash_command(), interactive=True, tty=True,
                                                                       user=args.user)
            command = get_tmux_session_command(command, session_name, script_name, default_command=default_command,
                                               keep_pane=True, detached=True)

        # execute command on the host OS
        print("\n Running below script on instance. \n")
        print(script_content)
        instance_manager.exec(command)
        update_run_status(args.run_id, RUNNING)



import boto3
from botocore.exceptions import ClientError, WaiterError
from spotml.commands.writers.abstract_output_writrer import AbstractOutputWriter
from spotml.providers.aws.cfn_templates.instance_profile.template import prepare_instance_profile_template
from spotml.providers.aws.resources.stack import Stack


class InstanceProfileStackManager(object):

    def __init__(self, project_name: str, instance_name: str, region: str):
        self._cf = boto3.client('cloudformation', region_name=region)
        self._region = region
        self._stack_name = 'spotml-instance-profile-%s-%s' % (project_name.lower(), instance_name.lower())

    def create_or_update_stack(self, managed_policy_arns: list, output: AbstractOutputWriter):
        """Creates or updates an instance profile.
        It was moved to a separate stack because creating of an instance profile resource takes 2 minutes.
        """
        # check that policies exist
        iam = boto3.client('iam', region_name=self._region)
        for policy_arn in managed_policy_arns:
            # if the policy doesn't exist, an error will be raised
            iam.get_policy(PolicyArn=policy_arn)

        template = prepare_instance_profile_template(managed_policy_arns)

        stack = Stack.get_by_name(self._cf, self._stack_name)
        try:
            if stack:
                # update the stack and wait until it will be updated
                self._update_stack(template, output)
            else:
                # create the stack and wait until it will be created
                self._create_stack(template, output)

            stack = Stack.get_by_name(self._cf, self._stack_name)
        except WaiterError:
            stack = None

        if not stack or stack.status not in ['CREATE_COMPLETE', 'UPDATE_COMPLETE']:
            raise ValueError('Stack "%s" was not created.\n'
                             'Please, see CloudFormation logs for the details.' % self._stack_name)

        profile_arn = [row['OutputValue'] for row in stack.outputs if row['OutputKey'] == 'ProfileArn'][0]

        return profile_arn

    def _create_stack(self, template: str, output: AbstractOutputWriter):
        """Creates the stack and waits until it will be created."""
        output.write('Creating IAM role for the instance...')

        stack = Stack.create_stack(
            cf=self._cf,
            StackName=self._stack_name,
            TemplateBody=template,
            Capabilities=['CAPABILITY_IAM'],
            OnFailure='DELETE',
        )

        # wait for the stack to be created
        stack.wait_stack_created(delay=15)

    def _update_stack(self, template: str, output: AbstractOutputWriter):
        """Updates the stack and waits until it will be updated."""
        try:
            updated_stack = Stack.update_stack(
                cf=self._cf,
                StackName=self._stack_name,
                TemplateBody=template,
                Capabilities=['CAPABILITY_IAM'],
            )
        except ClientError as e:
            # the stack was not updated because there are no changes
            updated_stack = None
            error_code = e.response.get('Error', {}).get('Code', 'Unknown')
            if error_code != 'ValidationError':
                raise e

        if updated_stack:
            # wait for the stack to be updated
            output.write('Updating IAM role for the instance...')
            updated_stack.wait_stack_updated(delay=15)



from abc import ABC


class AbstractInstance(ABC):

    @property
    def public_ip_address(self):
        raise NotImplementedError

    @property
    def is_running(self):
        """Returns true if the instance is running."""
        raise NotImplementedError

    @property
    def is_stopped(self):
        """Returns true if the instance is stopped, so it can be restarted."""
        raise NotImplementedError

    def terminate(self, wait: bool = True):
        raise NotImplementedError

    def stop(self, wait: bool = True):
        raise NotImplementedError



import logging
import os
from datetime import datetime

import boto3 as boto3
import requests as requests
from tabulate import tabulate
from dateutil import parser
from tzlocal import get_localzone

from spotml.config.config_utils import DEFAULT_CONFIG_FILENAME
from spotml.config.url import API_URL
from spotml.constants.constants import GENERIC_ERROR_MESSAGE


def get_local_time_from_string(utc_date_time_string):
    return parser.parse(utc_date_time_string).astimezone(get_localzone()).strftime('%b %d, %Y -%l:%M%p %Z')


def update_run_status(run_id, status):
    logging.info(f'Updating Run status to: {status}')
    r = requests.patch(f'{API_URL}/run/{run_id}',
                       data={'status': status, 'aws_batch_job_id': os.getenv('AWS_BATCH_JOB_ID')})
    if r.status_code != 201:
        print("Oops, looks like something went wrong in scheduling the run")
        print(GENERIC_ERROR_MESSAGE)
        raise Exception('Something went wrong updating Run Status', r)


def get_run(run_id):
    r = requests.get(f'{API_URL}/run/{run_id}')
    if r.status_code != 200:
        raise RuntimeError('Unable to get run metadata.')
    return r.json()


def print_run_status(should_print_logs: bool):
    session = boto3.session.Session()
    credentials = session.get_credentials()
    region = session.region_name
    access_key = credentials.access_key
    secret_key = credentials.secret_key

    with open(DEFAULT_CONFIG_FILENAME, 'r') as spotYml:
        files = {'spotml_config': spotYml}
        r = requests.get(f'{API_URL}/run/status', files=files,
                         data={'region': region, 'access_key': access_key,
                               'secret_key': secret_key})
        if r.status_code == 200:
            runs = r.json()['runs']
            multiple_job_logs = r.json()['logs']
            if len(runs) > 0:
                table = []
                for run in runs:
                    table.append(
                        [run['id'], run['script_name'], run['status'], get_local_time_from_string(run['created_date'])])
                print(tabulate(table, headers=['Run Id', 'Command', 'Status', 'Start Time'], tablefmt="psql"))
            else:
                print("No Active Runs")

            if should_print_logs:
                print_run_logs(multiple_job_logs)

        else:
            print("Sorry, Failed to get latest status runs.")
            print(GENERIC_ERROR_MESSAGE)


def print_run_logs(multiple_job_logs):
    print("\nLAST RUN LOGS: ")
    if len(multiple_job_logs) > 0:
        for job_log in multiple_job_logs:
            table = []
            for log in job_log:
                table.append(
                    [datetime.fromtimestamp(log['timestamp'] / 1000).strftime('%b %d, %Y -%l:%M%p %Z'),
                     log['message']])
            print(tabulate(table, headers=['Time', 'Log Text'], tablefmt="psql"))
    else:
        print("Nothing so far... \n")


def track_instance_start(ssh_key_path):
    session = boto3.session.Session()
    credentials = session.get_credentials()
    region = session.region_name
    access_key = credentials.access_key
    secret_key = credentials.secret_key

    with open(ssh_key_path, 'r') as sshKey:
        with open(DEFAULT_CONFIG_FILENAME, 'r') as spotYml:
            files = {'spotml_config': spotYml, 'ssh_key': sshKey}
            r = requests.post(f'{API_URL}/run/track-instance-start', files=files,
                              data={'region': region, 'access_key': access_key,
                                    'secret_key': secret_key})
            if r.status_code == 200:
                old_runs = r.json()['old_runs']
                print_old_runs(old_runs)
                print("\nTracking instance uptime now.\n")
            else:
                print("Failed to stop existing runs. Use \"spotml status%s\" command to check the status of runs.")
                print("\nOops, Unable to start track instance uptime.\n")
                print(GENERIC_ERROR_MESSAGE)


def track_instance_stop():
    session = boto3.session.Session()
    credentials = session.get_credentials()
    region = session.region_name
    access_key = credentials.access_key
    secret_key = credentials.secret_key
    with open(DEFAULT_CONFIG_FILENAME, 'r') as spotYml:
        files = {'spotml_config': spotYml}
        r = requests.post(f'{API_URL}/run/track-instance-stop', files=files,
                          data={'region': region, 'access_key': access_key,
                                'secret_key': secret_key})

        if r.status_code == 200:
            old_runs = r.json()['old_runs']
            print_old_runs(old_runs)
            print("\nStopped tracking instance uptime.")
        else:
            print("\nOops, Unable to stop track instance uptime.")
            print(GENERIC_ERROR_MESSAGE)


def get_instance_tracking_status():
    session = boto3.session.Session()
    credentials = session.get_credentials()
    region = session.region_name
    access_key = credentials.access_key
    secret_key = credentials.secret_key
    with open(DEFAULT_CONFIG_FILENAME, 'r') as spotYml:
        files = {'spotml_config': spotYml}
        r = requests.get(f'{API_URL}/instance/tracking-status', files=files,
                         data={'region': region, 'access_key': access_key,
                               'secret_key': secret_key})

        if r.status_code == 200:
            return r.json()['is_tracking_idle_time']
        else:
            return False


def print_old_runs(old_runs):
    if len(old_runs) > 0:
        print("\nStopped managing all active runs.")
        print("STOPPED RUNS")
        table = []
        for run in old_runs:
            table.append(
                [run['id'], run['script_name'], run['status'], get_local_time_from_string(run['created_date'])])
        print(tabulate(table, headers=['Run Id', 'Command', 'Status', 'Start Time'], tablefmt="psql"))
    else:
        print("\nNo active runs found.")



import boto3
from spotml.commands.writers.abstract_output_writrer import AbstractOutputWriter
from spotml.deployment.abstract_cloud_instance.abstract_instance_deployment import AbstractInstanceDeployment
from spotml.deployment.container.docker.docker_commands import DockerCommands
from spotml.exceptions.SpotCapacityException import SpotCapacityException
from spotml.providers.aws.cfn_templates.instance.template import prepare_instance_template, get_template_parameters
from spotml.providers.aws.data_transfer import DataTransfer
from spotml.providers.aws.helpers.availability_zone import update_availability_zone
from spotml.providers.aws.helpers.instance_prices import check_max_spot_price
from spotml.providers.aws.helpers.subnet import check_az_and_subnet
from spotml.providers.aws.resource_managers.key_pair_manager import KeyPairManager
from spotml.deployment.utils.print_info import render_volumes_info_table
from spotml.providers.aws.resources.instance import Instance
from spotml.providers.aws.config.instance_config import InstanceConfig
from spotml.providers.aws.deletion_policies import apply_deletion_policies
from spotml.providers.aws.resource_managers.instance_profile_stack_manager import InstanceProfileStackManager
from spotml.providers.aws.helpers.logs import download_logs
from spotml.providers.aws.resource_managers.instance_stack_manager import InstanceStackManager


class InstanceDeployment(AbstractInstanceDeployment):
    instance_config: InstanceConfig

    def __init__(self, instance_config: InstanceConfig):
        super().__init__(instance_config)

        self._project_name = instance_config.project_config.project_name
        self._ec2 = boto3.client('ec2', region_name=instance_config.region)

    @property
    def stack_manager(self) -> InstanceStackManager:
        return InstanceStackManager(self._project_name, self.instance_config.name, self.instance_config.region)

    @property
    def key_pair_manager(self) -> KeyPairManager:
        return KeyPairManager(self._ec2, self._project_name, self.instance_config.region)

    def get_instance(self) -> Instance:
        return Instance.get_by_stack_name(self._ec2, self.stack_manager.name)

    def maybe_create_key(self) -> None:
        self.key_pair_manager.maybe_create_key()

    def deploy(self, container_commands: DockerCommands, bucket_name: str,
               data_transfer: DataTransfer, output: AbstractOutputWriter, dry_run: bool = False):
        # get deployment availability zone
        availability_zone = update_availability_zone(self._ec2, self.instance_config.availability_zone,
                                                     self.instance_config.volumes)

        # check availability zone and subnet configuration
        check_az_and_subnet(self._ec2, self.instance_config.region, availability_zone, self.instance_config.subnet_id)

        # check the maximum price for a spot instance
        check_max_spot_price(self._ec2, self.instance_config.instance_type, self.instance_config.is_spot_instance,
                             self.instance_config.max_price, availability_zone)

        # sync the project with the S3 bucket - removing this since this is now handled in run and start separately.
        # if bucket_name is not None:
        #     output.write('Syncing the project with the S3 bucket...')
        #     data_transfer.upload_local_to_bucket(bucket_name, dry_run=dry_run)

        # create or update instance profile
        if not dry_run:
            instance_profile_stack_manager = InstanceProfileStackManager(
                self._project_name, self.instance_config.name, self.instance_config.region)
            if not self.instance_config.instance_profile_arn:
                instance_profile_arn = instance_profile_stack_manager.create_or_update_stack(
                    self.instance_config.managed_policy_arns, output=output)
            else:
                instance_profile_arn = self.instance_config.instance_profile_arn
        else:
            instance_profile_arn = None

        # create a key pair if it doesn't exist
        if not dry_run:
            self.maybe_create_key()

        output.write('Preparing CloudFormation template...')

        # prepare CloudFormation template
        with output.prefix('  '):
            template = prepare_instance_template(
                ec2=self._ec2,
                instance_config=self.instance_config,
                docker_commands=container_commands,
                availability_zone=availability_zone,
                sync_project_cmd=data_transfer.get_download_bucket_to_instance_command(bucket_name=bucket_name),
                output=output,
            )

            # get parameters for the template
            parameters = get_template_parameters(
                ec2=self._ec2,
                instance_config=self.instance_config,
                instance_profile_arn=instance_profile_arn,
                bucket_name=bucket_name,
                key_pair_name=self.key_pair_manager.key_name,
                output=output,
            )

        # print information about the volumes
        output.write('\nVolumes:\n%s\n'
                     % render_volumes_info_table(self.instance_config.volume_mounts, self.instance_config.volumes))

        # create stack
        if not dry_run:
            stack = self.stack_manager.create_or_update_stack(template, parameters, self.instance_config, output)
            if stack.status != 'CREATE_COMPLETE':
                logs_str = 'Please, see CloudFormation logs for more details.'
                if stack.error:
                    logs_str += f'\n {stack.error} \n'

                # download CloudFormation logs from the instance if it was created
                if self.get_instance():
                    log_paths = download_logs(
                        bucket_name=bucket_name,
                        instance_name=self.instance_config.name,
                        stack_uuid=stack.stack_uuid,
                        region=self.instance_config.region,
                    )

                    logs_str = 'Please, see the logs for the details:\n  '
                    logs_str += '\n  '.join(log_paths)

                if stack.error and (
                        "not supported in your requested availability zone" in stack.error.lower() or
                        "no spot capacity available" in stack.error.lower() or
                        "insufficientinstancecapacity" in stack.error.lower()
                ):
                    raise SpotCapacityException('Stack "%s" was not created.\n%s' % (stack.name, logs_str))
                else:
                    raise ValueError('Stack "%s" was not created.\n%s' % (stack.name, logs_str))

    def delete(self, output: AbstractOutputWriter):
        # terminate the instance
        instance = self.get_instance()
        if instance:
            output.write('Terminating the instance... ', newline=False)
            instance.terminate()
            output.write('DONE')
        else:
            output.write('The instance was already terminated.')

        # delete the stack in background if it exists
        self.stack_manager.delete_stack(output, no_wait=True)

        output.write('Applying deletion policies for the volumes...')

        # apply deletion policies for the volumes
        with output.prefix('  '):
            apply_deletion_policies(self._ec2, self.instance_config.volumes, output)



import base64
import os
import shlex
import time
from spotml.deployment.utils.cli import shlex_join


def get_bash_command() -> str:
    return '/usr/bin/env bash'


def get_script_command(script_name: str, script_content: str, script_args: list = None,
                       logging: bool = False) -> str:
    """Encodes a multi-line script into base64 and returns a one-line command
    that unpacks the script to a temporary file and runs it."""

    # encode the script content to base64
    script_base64 = base64.b64encode(script_content.encode('utf-8')).decode('utf-8')

    # command to decode the script, save it to a temporary file and run inside the container
    script_args = shlex_join(script_args) if script_args else ''

    script_cmd = ' && '.join([
        'TMPDIR=${TMPDIR%/}',
        'TMP_SCRIPT_PATH=$(mktemp ${TMPDIR:-/tmp}/spotml-%s.XXXXXXXX)' % script_name,
        'chmod +x $TMP_SCRIPT_PATH',
        'echo %s | base64 -d > $TMP_SCRIPT_PATH' % script_base64,
        '$TMP_SCRIPT_PATH ' + script_args,
    ])

    # log the command output to a file
    if logging:
        log_file_path = '/var/log/spotml/run/%s-%d.log' % (script_name, time.time())
        script_cmd = get_log_command(script_cmd, log_file_path)

    # execute the command with bash
    script_cmd = '%s -c %s ' % (get_bash_command(), shlex.quote(script_cmd))

    return script_cmd


def get_send_keys_command(script_content: str, session_name: str) -> str:
    """Encodes a multi-line script into base64 and returns a one-line command
    that unpacks the script to a temporary file and runs it."""

    print('Sending below command to be executed:')
    print(script_content)

    # encode the script content to base64
    script_base64 = base64.b64encode(script_content.encode('utf-8')).decode('utf-8')

    send_keys_cmd = f';tmux send-keys -t {session_name} C-c; tmux send-keys -t {session_name} "echo {script_base64} | base64 -d | bash" ENTER'
    return send_keys_cmd


def get_log_command(command: str, log_file_path: str) -> str:
    # log the command outputs to a file on the host OS
    log_dir = os.path.dirname(log_file_path)
    log_cmd = '; '.join([
        'set -o pipefail',
        ' && '.join([
            'mkdir -pm 777 ' + shlex.quote(log_dir),
            '(%s) 2>&1 | tee %s' % (command, shlex.quote(log_file_path)),
        ]),
    ])

    return log_cmd


def get_tmux_session_command(command: str, session_name: str, window_name: str = None, default_command: str = None,
                             keep_pane: bool = False, detached: bool = False, send_keys_command: str = None) -> str:
    if detached:
        session_cmd = 'tmux kill-session -t ' + session_name + ';tmux new -d -s ' + session_name
    else:
        session_cmd = 'tmux new -A -s ' + session_name

    if window_name:
        session_cmd += ' -n ' + window_name

    if command:
        # keep the pane alive when the script is finished
        keep_pane_cmd = 'tmux set -w remain-on-exit on; ' if keep_pane else ''

        # set the default command (to automatically run bash inside the container when a new window is created)
        default_command_cmd = ('tmux set default-command %s; ' % shlex.quote(default_command)) \
            if default_command else ''

        # keep the pane alive if the script is failed
        tmux_cmd = '%s%s(%s) || tmux set -w remain-on-exit on' % (keep_pane_cmd, default_command_cmd, command)

        # run the command inside the tmux session
        session_cmd += ' ' + shlex.quote(tmux_cmd)
        if send_keys_command:
            session_cmd += send_keys_command

    # use tmux only if it's installed
    session_cmd = 'if command -v tmux &> /dev/null; then %s; else %s; fi' % (session_cmd, command)

    return session_cmd


def get_ssh_command(host: str, port: int, user: str, key_path: str, command: str, env_vars: dict = None,
                    tty: bool = True, quiet: bool = False) -> str:
    ssh_command = 'ssh -i %s -o StrictHostKeyChecking=no -o ConnectTimeout=10' % shlex.quote(key_path)

    if tty:
        ssh_command += ' -t'

    if port != 22:
        ssh_command += ' -p %d' % port

    if quiet:
        ssh_command += ' -q'

    # export environmental variables
    if env_vars:
        export_cmd = '; '.join(['export %s=%s' % (name, shlex.quote(val)) for name, val in env_vars.items()])
        command = '%s; %s' % (export_cmd, command)

    # final SSH command
    ssh_command += ' %s@%s %s' % (user, host, shlex.quote(command))

    return ssh_command



import fnmatch
from shutil import which
import os
from typing import List

from spotml.deployment.utils.cli import shlex_join


def check_gsutil_installed():
    """Checks that gsutil is installed."""
    if which('gsutil') is None:
        raise ValueError('gsutil is not installed.')


def get_rsync_command(from_path: str, to_path: str, filters: List[dict] = None, delete: bool = False,
                      quiet: bool = False, dry_run: bool = False):
    args = ['gsutil', '-m']
    if quiet:
        args.append('-q')

    args += ['rsync', '-r']

    if filters:
        if (len(filters) > 1) or (len(filters[0]) > 1) or ('include' in filters[0]):
            raise ValueError('At the moment GCP provider supports only one list of exclude filters.')

        path_regs = []
        for path in filters[0]['exclude']:
            path = path.replace('/', os.sep)  # fix for Windows machines
            path_regs.append(fnmatch.translate(path)[4:-3])

        filter_regex = '^(%s)$' % '|'.join(path_regs)
        args += ['-x', filter_regex]

    if delete:
        args.append('-d')

    if dry_run:
        args.append('-n')

    args += [from_path, to_path]

    return shlex_join(args)



from spotml.deployment.abstract_cloud_instance.resources.abstract_bucket import AbstractBucket


class Bucket(AbstractBucket):

    def __init__(self, data: dict):
        self._data = data

    @property
    def name(self) -> str:
        return self._data['Name']



from typing import List
from spotml.commands.writers.abstract_output_writrer import AbstractOutputWriter
from spotml.config.abstract_instance_volume import AbstractInstanceVolume
from spotml.providers.gcp.config.disk_volume import DiskVolume
from spotml.providers.gcp.helpers.ce_client import CEClient
from spotml.providers.gcp.resources.disk import Disk
from spotml.providers.gcp.resources.snapshot import Snapshot


def create_disks(ce: CEClient, volumes: List[AbstractInstanceVolume], output: AbstractOutputWriter,
                 dry_run: bool = False):
    disks_to_create = []

    # do some checks and prepare disk parameters
    for i, volume in enumerate(volumes):
        if isinstance(volume, DiskVolume):
            # check if the disk already exists
            disk = Disk.get_by_name(ce, volume.disk_name)
            if disk:
                # check if the volume is available
                if not disk.is_available():
                    raise ValueError('Disk "%s" is not available (status: %s).'
                                     % (volume.disk_name, disk.status))

                # check size of the volume
                if volume.size and (volume.size != disk.size):
                    raise ValueError('Specified size for the "%s" volume (%dGB) doesn\'t match the size of the '
                                     'existing disk (%dGB).' % (volume.name, volume.size, disk.size))

                output.write('- disk "%s" will be attached' % disk.name)
            else:
                # check if the snapshot exists
                snapshot = Snapshot.get_by_name(ce, volume.disk_name)
                if snapshot:
                    # disk will be restored from the snapshot
                    # check size of the volume
                    if volume.size and (volume.size < snapshot.size):
                        raise ValueError('Specified size for the "%s" volume (%dGB) is less than size of the '
                                         'snapshot (%dGB).'
                                         % (volume.name, volume.size, snapshot.size))

                    output.write('- disk "%s" will be restored from the snapshot' % volume.disk_name)

                    disks_to_create.append((volume.disk_name, volume.size, snapshot.self_link))
                else:
                    # empty volume will be created, check that the size is specified
                    if not volume.size:
                        raise ValueError('Size for the new disk is required.')

                    if volume.size < 10:
                        raise ValueError('Size of a disk cannot be less than 10GB.')

                    disks_to_create.append((volume.disk_name, volume.size, None))

    # create disks
    for disk_name, disk_size, snapshot_link in disks_to_create:
        if not dry_run:
            ce.create_disk(disk_name, disk_size, snapshot_link)

        output.write('- disk "%s" was created' % disk_name)



import boto3
from spotml.commands.writers.abstract_output_writrer import AbstractOutputWriter
from spotml.providers.aws.resources.stack import Stack, Task
from spotml.providers.aws.config.instance_config import InstanceConfig


class InstanceStackManager(object):

    def __init__(self, project_name: str, instance_name: str, region: str):
        self._cf = boto3.client('cloudformation', region_name=region)
        self._ec2 = boto3.client('ec2', region_name=region)
        self._region = region
        self._stack_name = 'spotml-instance-%s-%s' % (project_name.lower(), instance_name.lower())

    @property
    def name(self):
        return self._stack_name

    def create_or_update_stack(self, template: str, parameters: dict, instance_config: InstanceConfig,
                               output: AbstractOutputWriter):
        """Runs CloudFormation template."""

        # delete the stack if it exists
        stack = Stack.get_by_name(self._cf, self._stack_name)
        if stack:
            self.delete_stack(output)

        # create new stack
        stack = Stack.create_stack(
            cf=self._cf,
            StackName=self._stack_name,
            TemplateBody=template,
            Parameters=[{'ParameterKey': key, 'ParameterValue': value} for key, value in parameters.items()],
            Capabilities=['CAPABILITY_IAM'],
            OnFailure='DO_NOTHING',
        )

        output.write('Waiting for the stack to be created...')

        tasks = [
            Task(
                message='launching the instance',
                start_resource=None,
                finish_resource='Instance',
                enabled=True,
            ),
            Task(
                message='preparing the instance',
                start_resource='Instance',
                finish_resource='MountingVolumesSignal',
                enabled=True,
            ),
            Task(
                message='mounting volumes',
                start_resource='MountingVolumesSignal',
                finish_resource='SettingDockerRootSignal',
                enabled=bool(instance_config.volumes),
            ),
            Task(
                message='setting Docker data root',
                start_resource='SettingDockerRootSignal',
                finish_resource='SyncingProjectSignal',
                enabled=bool(instance_config.docker_data_root),
            ),
            Task(
                message='syncing project files',
                start_resource='SyncingProjectSignal',
                finish_resource='RunningInstanceStartupCommandsSignal',
                enabled=True,
            ),
            Task(
                message='running instance startup commands',
                start_resource='RunningInstanceStartupCommandsSignal',
                finish_resource='BuildingDockerImageSignal',
                enabled=bool(instance_config.commands),
            ),
            Task(
                message='building Docker image',
                start_resource='BuildingDockerImageSignal',
                finish_resource='StartingContainerSignal',
                enabled=bool(instance_config.dockerfile_path),
            ),
            Task(
                message='starting container',
                start_resource='StartingContainerSignal',
                finish_resource='RunningContainerStartupCommandsSignal',
                enabled=True,
            ),
            Task(
                message='running container startup commands',
                start_resource='RunningContainerStartupCommandsSignal',
                finish_resource='DockerReadyWaitCondition',
                enabled=bool(instance_config.container_config.commands),
            ),
        ]

        # wait for the stack to be created
        with output.prefix('  '):
            stack.wait_tasks(tasks, resource_success_status='CREATE_COMPLETE', resource_fail_status='CREATE_FAILED',
                             output=output)
            stack = stack.wait_status_changed(stack_waiting_status='CREATE_IN_PROGRESS', output=output)

        return stack

    def delete_stack(self, output: AbstractOutputWriter, no_wait=False):
        stack = Stack.get_by_name(self._cf, self._stack_name)
        if not stack:
            return

        if not no_wait:
            output.write('Waiting for the stack to be deleted...')

        # delete the stack
        try:
            stack.delete()
            if not no_wait:
                stack.wait_stack_deleted()
        except Exception as e:
            raise ValueError('Stack "%s" was not deleted. Error: %s\n'
                             'See CloudFormation logs for details.' % (self._stack_name, str(e)))



#!/usr/bin/env python

import sys
import logging
import spotml
from spotml.parser import get_parser
from spotml.commands.writers.output_writrer import OutputWriter


def run_command(output: OutputWriter):
    parser = get_parser()
    args = sys.argv[1:]
    # display the version
    if '-V' in args:
        output.write(spotml.__version__)
        sys.exit(0)
    # separate Spotml arguments from custom arguments
    custom_args = []
    if '--' in args:
        dd_idx = args.index('--')
        custom_args = args[(dd_idx + 1):]
        args = args[:dd_idx]
    # parse arguments
    args = parser.parse_args(args)
    args.custom_args = custom_args
    # logging
    logging_level = logging.DEBUG if 'debug' in args and args.debug else logging.WARNING
    logging.basicConfig(level=logging_level, format='[%(levelname)s] %(message)s')
    if 'command' not in args:
        parser.print_help()
        sys.exit(1)
    args.command.run(args, output)


def main():
    try:
        output = OutputWriter()
        run_command(output)
    except Exception as e:
        output.write('Error:\n------\n%s' % str(e))
        sys.exit(1)


if __name__ == '__main__':
    main()



from spotml.commands.writers.abstract_output_writrer import AbstractOutputWriter
from spotml.deployment.abstract_cloud_instance.abstract_instance_deployment import AbstractInstanceDeployment
from spotml.deployment.container.docker.docker_commands import DockerCommands
from spotml.deployment.utils.print_info import render_volumes_info_table
from spotml.providers.gcp.config.instance_config import InstanceConfig
from spotml.providers.gcp.data_transfer import DataTransfer
from spotml.providers.gcp.dm_templates.instance.instance_template import prepare_instance_template
from spotml.providers.gcp.helpers.image import get_image
from spotml.providers.gcp.helpers.volumes import create_disks
from spotml.providers.gcp.resource_managers.instance_stack_manager import InstanceStackManager
from spotml.providers.gcp.helpers.ce_client import CEClient
from spotml.providers.gcp.helpers.gcp_credentials import GcpCredentials
from spotml.providers.gcp.resource_managers.ssh_key_manager import SshKeyManager
from spotml.providers.gcp.resources.instance import Instance
from spotml.providers.gcp.helpers.deployment import check_gpu_configuration


class InstanceDeployment(AbstractInstanceDeployment):

    instance_config: InstanceConfig

    def __init__(self, instance_config: InstanceConfig):
        super().__init__(instance_config)

        self._project_name = instance_config.project_config.project_name
        self._credentials = GcpCredentials()
        self._ce = CEClient(self._credentials.project_id, instance_config.zone)

    @property
    def stack_manager(self) -> InstanceStackManager:
        return InstanceStackManager(self.instance_config.machine_name, self._credentials.project_id, self.instance_config.zone)

    @property
    def ssh_key_manager(self) -> SshKeyManager:
        return SshKeyManager(self._project_name, self.instance_config.zone)

    def get_instance(self) -> Instance:
        return Instance.get_by_name(self._ce, self.instance_config.machine_name)

    def deploy(self, container_commands: DockerCommands, bucket_name: str,
               data_transfer: DataTransfer, output: AbstractOutputWriter, dry_run: bool = False):
        # check machine type
        if not self._ce.get_machine_types(self.instance_config.machine_type):
            raise ValueError('"%s" machine type is not available in the "%s" zone.'
                             % (self.instance_config.machine_type, self.instance_config.zone))

        # check GPU configuration
        check_gpu_configuration(self._ce, self.instance_config.gpu)

        # remove the stack it it exists to make all the disks available
        stack_manager = self.stack_manager
        stack_manager.delete_stack(output=output)

        # sync the project with the S3 bucket
        if bucket_name is not None:
            output.write('Syncing the project with the bucket...')
            data_transfer.upload_local_to_bucket(bucket_name, dry_run=dry_run)

        # create volumes
        if self.instance_config.volumes:
            # create disks
            output.write('\nCreating disks...')
            with output.prefix('  '):
                create_disks(self._ce, self.instance_config.volumes, output=output, dry_run=dry_run)
            output.write('')

        # prepare Deployment Manager template
        output.write('Preparing the deployment template...')
        with output.prefix('  '):
            # get an image
            image_link = get_image(self._ce, self.instance_config.image_uri, self.instance_config.image_name).self_link

            # get or create an SSH key
            public_key_value = self.ssh_key_manager.get_public_key_value()

            # prepare the deployment template
            sync_project_cmd = data_transfer.get_download_bucket_to_instance_command(bucket_name=bucket_name)
            template = prepare_instance_template(
                instance_config=self.instance_config,
                docker_commands=container_commands,
                image_link=image_link,
                bucket_name=bucket_name,
                sync_project_cmd=sync_project_cmd,
                public_key_value=public_key_value,
                service_account_email=self._credentials.service_account_email,
                output=output,
            )

        output.write('')

        # print information about the volumes
        output.write('Volumes:\n%s\n' % render_volumes_info_table(self.instance_config.volume_mounts,
                                                                  self.instance_config.volumes))

        # create stack
        if not dry_run:
            stack_manager.create_stack(template, output=output)

    def delete(self, output: AbstractOutputWriter):
        self.stack_manager.delete_stack(output)

        # TODO: apply deletion policies



from typing import List
from spotml.commands.writers.abstract_output_writrer import AbstractOutputWriter
from spotml.config.abstract_instance_volume import AbstractInstanceVolume
from spotml.providers.aws.resources.snapshot import Snapshot
from spotml.providers.aws.resources.volume import Volume
from spotml.providers.aws.config.ebs_volume import EbsVolume


def apply_deletion_policies(ec2, volumes: List[AbstractInstanceVolume], output: AbstractOutputWriter):
    """Applies deletion policies to the EBS volumes."""

    # get volumes
    ebs_volumes = [volume for volume in volumes if isinstance(volume, EbsVolume)]

    # no volumes
    if not ebs_volumes:
        output.write('- no EBS volumes configured')
        return

    # apply deletion policies
    wait_snapshots = []
    for volume in ebs_volumes:
        # get EC2 volume
        try:
            ec2_volume = Volume.get_by_name(ec2, volume.ec2_volume_name)
        except Exception as e:
            output.write('- volume "%s" not found. Error: %s' % (volume.ec2_volume_name, str(e)))
            continue

        if not ec2_volume:
            output.write('- volume "%s" not found' % volume.ec2_volume_name)
            continue

        if not ec2_volume.is_available():
            output.write('- volume "%s" is not available (state: %s)'
                         % (volume.ec2_volume_name, ec2_volume.state))
            continue

        # apply deletion policies
        if volume.deletion_policy == EbsVolume.DP_RETAIN:
            # do nothing
            output.write('- volume "%s" is retained' % ec2_volume.name)

        elif volume.deletion_policy == EbsVolume.DP_DELETE:
            # delete EBS volume
            _delete_ec2_volume(ec2_volume, output)

        elif volume.deletion_policy == EbsVolume.DP_CREATE_SNAPSHOT \
                or volume.deletion_policy == EbsVolume.DP_UPDATE_SNAPSHOT:
            try:
                # rename a previous snapshot
                prev_snapshot = Snapshot.get_by_name(ec2, volume.ec2_volume_name)
                if prev_snapshot:
                    prev_snapshot.rename('%s-%d' % (prev_snapshot.name, prev_snapshot.creation_time))

                output.write('- creating a snapshot for the volume "%s"...' % ec2_volume.name)

                # create a new snapshot
                new_snapshot = ec2_volume.create_snapshot()

                # delete the EBS volume and a previous snapshot only after a new snapshot will be created
                wait_snapshots.append({
                    'new_snapshot': new_snapshot,
                    'prev_snapshot': prev_snapshot,
                    'ec2_volume': ec2_volume,
                    'deletion_policy': volume.deletion_policy,
                })
            except Exception as e:
                output.write('- snapshot for the volume "%s" was not created. Error: %s'
                             % (volume.ec2_volume_name, str(e)))

        else:
            raise ValueError('Unsupported deletion policy: "%s".' % volume.deletion_policy)

    # wait until all snapshots will be created
    for resources in wait_snapshots:
        try:
            resources['new_snapshot'].wait_snapshot_completed()
            output.write('- snapshot for the volume "%s" was created' % resources['new_snapshot'].name)
        except Exception as e:
            output.write('- snapshot "%s" was not created. Error: %s' % (resources['new_snapshot'].name, str(e)))
            continue

        # delete a previous snapshot if it's the "update_snapshot" deletion policy
        if (resources['deletion_policy'] == EbsVolume.DP_UPDATE_SNAPSHOT) and resources['prev_snapshot']:
            _delete_snapshot(resources['prev_snapshot'], output)

        # delete the EBS volume
        _delete_ec2_volume(resources['ec2_volume'], output)


def _delete_ec2_volume(ec2_volume: Volume, output: AbstractOutputWriter):
    try:
        ec2_volume.delete()
        output.write('- volume "%s" was deleted' % ec2_volume.name)
    except Exception as e:
        output.write('- volume "%s" was not deleted. Error: %s' % (ec2_volume.name, str(e)))


def _delete_snapshot(snapshot: Snapshot, output: AbstractOutputWriter):
    try:
        snapshot.delete()
        output.write('- previous snapshot "%s" was deleted' % snapshot.name)
    except Exception as e:
        output.write('- previous snapshot "%s" was not deleted. Error: %s' % (snapshot.name, str(e)))



from spotml.commands.abstract_provider_command import AbstractProviderCommand
from spotml.providers.aws.commands.clean_logs import CleanLogsCommand
from spotml.providers.aws.commands.spot_prices import SpotPricesCommand


class AwsCommand(AbstractProviderCommand):

    name = 'aws'
    description = 'AWS commands'
    commands = [
        SpotPricesCommand,
        CleanLogsCommand,
    ]



import os
from typing import List
import chevron
from spotml.commands.writers.abstract_output_writrer import AbstractOutputWriter
from spotml.config.tmp_dir_volume import TmpDirVolume
from spotml.config.validation import is_subdir
from spotml.config.abstract_instance_volume import AbstractInstanceVolume
from spotml.deployment.container.docker.docker_commands import DockerCommands
from spotml.deployment.container.docker.scripts.container_bash_script import ContainerBashScript
from spotml.deployment.container.docker.scripts.start_container_script import StartContainerScript
from spotml.deployment.abstract_cloud_instance.file_structure import CONTAINER_BASH_SCRIPT_PATH, \
    INSTANCE_STARTUP_SCRIPTS_DIR, CONTAINERS_TMP_DIR, INSTANCE_SPOTML_TMP_DIR
from spotml.providers.gcp.config.disk_volume import DiskVolume
from spotml.providers.gcp.config.instance_config import InstanceConfig


def prepare_instance_template(instance_config: InstanceConfig, docker_commands: DockerCommands, image_link: str,
                              bucket_name: str, sync_project_cmd: str, public_key_value: str,
                              service_account_email: str, output: AbstractOutputWriter):
    """Prepares deployment template to run an instance."""

    # get disk attachments
    disk_attachments, disk_device_names, disk_mount_dirs = \
        _get_disk_attachments(instance_config.volumes, instance_config.zone)

    # run sync command as a non-root user
    if instance_config.container_config.run_as_host_user:
        sync_project_cmd = 'sudo -u %s %s' % (instance_config.user, sync_project_cmd)

    startup_scripts_templates = [
        {
            'filename': '01_prepare_instance.sh',
            'params': {
                'CONTAINER_BASH_SCRIPT_PATH': CONTAINER_BASH_SCRIPT_PATH,
                'CONTAINER_BASH_SCRIPT': ContainerBashScript(docker_commands).render(),
                'IS_GPU_INSTANCE': bool(instance_config.gpu),
                'SSH_USERNAME': instance_config.user,
                'SPOTML_TMP_DIR': INSTANCE_SPOTML_TMP_DIR,
                'CONTAINERS_TMP_DIR': CONTAINERS_TMP_DIR,
            },
        },
        {
            'filename': '02_mount_volumes.sh',
            'params': {
                'DISK_DEVICE_NAMES': ('"%s"' % '" "'.join(disk_device_names)) if disk_device_names else '',
                'DISK_MOUNT_DIRS': ('"%s"' % '" "'.join(disk_mount_dirs)) if disk_mount_dirs else '',
                'TMP_VOLUME_DIRS': [{'PATH': volume.host_path} for volume in instance_config.volumes
                                    if isinstance(volume, TmpDirVolume)],
            },
        },
        {
            'filename': '03_set_docker_root.sh',
            'params': {
                'DOCKER_DATA_ROOT_DIR': instance_config.docker_data_root,
            },
        },
        {
            'filename': '04_sync_project.sh',
            'params': {
                'HOST_PROJECT_DIR': instance_config.host_project_dir,
                'SYNC_PROJECT_CMD': sync_project_cmd,
            },
        },
        {
            'filename': '05_run_instance_startup_commands.sh',
            'params': {
                'INSTANCE_STARTUP_SCRIPTS_DIR': INSTANCE_STARTUP_SCRIPTS_DIR,
                'INSTANCE_STARTUP_COMMANDS': instance_config.commands,
            },
        },
    ]

    # render startup scripts
    startup_scripts_content = []
    for template in startup_scripts_templates:
        with open(os.path.join(os.path.dirname(__file__), 'data', 'startup_scripts', template['filename'])) as f:
            content = f.read()

        startup_scripts_content.append({
            'filename': template['filename'],
            'content': chevron.render(content, template['params'])
        })

    startup_scripts_content.append({
        'filename': '06_start_container.sh',
        'content': StartContainerScript(docker_commands).render(print_trace=True),
    })

    # render the main startup script
    with open(os.path.join(os.path.dirname(__file__), 'data', 'startup_script.sh.tpl')) as f:
        startup_script = f.read()

    startup_script = chevron.render(startup_script, {
        'MACHINE_NAME': instance_config.machine_name,
        'INSTANCE_STARTUP_SCRIPTS_DIR': INSTANCE_STARTUP_SCRIPTS_DIR,
        'STARTUP_SCRIPTS': startup_scripts_content,
    })

    # render the template
    with open(os.path.join(os.path.dirname(__file__), 'data', 'template.yaml')) as f:
        template = f.read()

    template = chevron.render(template, {
        'SERVICE_ACCOUNT_EMAIL': service_account_email,
        'ZONE': instance_config.zone,
        'MACHINE_TYPE': instance_config.machine_type,
        'SOURCE_IMAGE': image_link,
        'BOOT_DISK_SIZE': instance_config.boot_disk_size,
        'MACHINE_NAME': instance_config.machine_name,
        'PREEMPTIBLE': 'true' if instance_config.is_preemptible_instance else 'false',
        'GPU_TYPE': instance_config.gpu['type'] if instance_config.gpu else '',
        'GPU_COUNT': instance_config.gpu['count'] if instance_config.gpu else 0,
        'DISK_ATTACHMENTS': disk_attachments,
        'SSH_USERNAME': instance_config.user,
        'PUB_KEY_VALUE': public_key_value,
        'PORTS': ', '.join([str(port) for port in set([22] + instance_config.ports)]),
    }, partials_dict={
        'STARTUP_SCRIPT': startup_script,
    })

    # print some information about the deployment
    output.write('- image URL: ' + '/'.join(image_link.split('/')[-5:]))
    output.write('- zone: ' + instance_config.zone)
    output.write('- preemptible VM' if instance_config.is_preemptible_instance else '- on-demand VM')
    output.write(('- GPUs: %d x %s' % (instance_config.gpu['count'], instance_config.gpu['type']))
                 if instance_config.gpu else '- no GPUs')

    # print name of the volume where Docker data will be stored
    if instance_config.docker_data_root:
        docker_data_volume_name = [volume.name for volume in instance_config.volumes
                                   if is_subdir(instance_config.docker_data_root, volume.host_path)][0]
        output.write('- Docker data will be stored on the "%s" volume' % docker_data_volume_name)

    return template


def _get_disk_attachments(volumes: List[AbstractInstanceVolume], zone: str):
    disk_attachments = []
    disk_device_names = []
    disk_mount_dirs = []

    for i, volume in enumerate(volumes):
        if isinstance(volume, DiskVolume):
            device_name = 'disk-%d' % (i + 1)
            disk_device_names.append(device_name)
            disk_mount_dirs.append(volume.mount_dir)
            disk_attachments.append({
                'DISK_LINK': 'zones/%s/disks/%s' % (zone, volume.disk_name),
                'DEVICE_NAME': device_name,
            })

    return disk_attachments, disk_device_names, disk_mount_dirs



import json
import googleapiclient.discovery
from googleapiclient.errors import HttpError


class DMClient(object):
    """Deployment Manager client."""

    def __init__(self, project_id: str, zone: str):
        self._project_id = project_id
        self._zone = zone
        self._client = googleapiclient.discovery.build('deploymentmanager', 'v2', cache_discovery=False)

    def get(self, deployment_name: str):
        try:
            res = self._client.deployments().get(project=self._project_id, deployment=deployment_name).execute()
        except HttpError as e:
            data = json.loads(e.content.decode('utf-8'))
            if data['error']['code'] != 404:
                raise e
            res = None

        return res

    def deploy(self, deployment_name: str, template: str, dry_run: bool = False):
        res = self._client.deployments().insert(project=self._project_id, body={
            'name': deployment_name,
            'target': {
                'config': {
                    'content': template,
                },
            },
        }, preview=dry_run).execute()

        return res

    def stop(self, deployment_name: str, fingerprint: str):
        res = self._client.deployments().stop(project=self._project_id, deployment=deployment_name, body={
            'fingerprint': fingerprint,
        }).execute()

        return res

    def delete(self, deployment_name: str):
        """Deletes a deployment and all of the resources in the deployment."""
        res = self._client.deployments().delete(project=self._project_id, deployment=deployment_name).execute()
        return res

    def get_resource(self, deployment_name: str, resource_name: str) -> dict:
        try:
            res = self._client.resources().get(project=self._project_id,
                                               deployment=deployment_name,
                                               resource=resource_name).execute()
        except HttpError as e:
            data = json.loads(e.content.decode('utf-8'))
            if data['error']['code'] != 404:
                raise e
            res = None

        return res



from argparse import Namespace, ArgumentParser

from spotml.commands.abstract_config_command import AbstractConfigCommand
from spotml.commands.writers.abstract_output_writrer import AbstractOutputWriter
from spotml.deployment.abstract_instance_manager import AbstractInstanceManager
from spotml.errors.instance_not_running import InstanceNotRunningError
from spotml.services.run_service import track_instance_start


class StartCommand(AbstractConfigCommand):
    name = 'start'
    description = 'Start an instance with a container'

    def configure(self, parser: ArgumentParser):
        super().configure(parser)
        parser.add_argument('-C', '--container', action='store_true', help='Starts or restarts container on the '
                                                                           'running instance')
        parser.add_argument('--dry-run', action='store_true', help='Displays the steps that would be performed '
                                                                   'using the specified command without actually '
                                                                   'running them')

    def _run(self, instance_manager: AbstractInstanceManager, args: Namespace, output: AbstractOutputWriter):
        dry_run = args.dry_run
        instance_manager.check_dockerfile_exists()
        instance_manager.create_or_get_bucket(output, False)
        instance_manager.sync_to_bucket(output, False)

        if args.container:
            # check that the instance is started
            if not instance_manager.is_running():
                raise InstanceNotRunningError(instance_manager.instance_config.name)

            # start a container on the running instance
            instance_manager.start_container(output, dry_run=dry_run)

            if not dry_run:
                instance_name = ''
                if len(instance_manager.project_config.instances) > 1:
                    instance_name = ' ' + instance_manager.instance_config.name

                output.write('\nContainer was successfully started.\n'
                             'Use the "spotml sh%s" command to connect to the container.\n'
                             % instance_name)
        else:
            # start the instance
            with output.prefix('[dry-run] ' if dry_run else ''):
                instance_manager.start(output, dry_run)
                track_instance_start(instance_manager.ssh_key_path)

            if not dry_run:
                instance_name = ''
                if len(instance_manager.project_config.instances) > 1:
                    instance_name = ' ' + instance_manager.instance_config.name

                output.write('\n%s\n'
                             '\nUse the "spotml sh%s" command to connect to the container.\n'
                             % (instance_manager.get_status_text(), instance_name))



MAX_IDLE_MINUTES = 5
INTERNAL_TAG = '[INTERNAL]:'
GENERIC_ERROR_MESSAGE = "Please message us on our slack channel or leave a note at bugs@spotml.io"
SPOT_STRATEGY_SPOT = 'spot'
SPOT_STRATEGY_ON_DEMAND = 'on-demand'



from typing import List
from spotml.config.abstract_instance_volume import AbstractInstanceVolume
from spotml.providers.aws.config.ebs_volume import EbsVolume
from spotml.providers.aws.resources.volume import Volume


def update_availability_zone(ec2, availability_zone: str, volumes: List[AbstractInstanceVolume]):
    """Checks that existing volumes located in the same AZ and the AZ from the
    config file matches volumes AZ.

    Args:
        ec2: EC2 boto3 client
        availability_zone: Availability Zone from the configuration.
        volumes: List of volume objects.

    Returns:
        The final AZ where the instance should be run or an empty string if
        the instance can be run in any AZ.

    Raises:
        ValueError: AZ in the config file doesn't match the AZs of the volumes or
            AZs of the volumes are different.
    """
    availability_zone = availability_zone
    for volume in volumes:
        if isinstance(volume, EbsVolume):
            ec2_volume = Volume.get_by_name(ec2, volume.ec2_volume_name)
            if ec2_volume:
                if availability_zone and (availability_zone != ec2_volume.availability_zone):
                    raise ValueError(
                        'The availability zone in the configuration file doesn\'t match the availability zone '
                        'of the existing volume or you have two existing volumes in different availability '
                        'zones.')

                # update availability zone
                availability_zone = ec2_volume.availability_zone

    return availability_zone



import re
from spotml.deployment.abstract_cloud_instance.abstract_bucket_manager import AbstractBucketManager
from spotml.deployment.abstract_cloud_instance.errors.bucket_not_found import BucketNotFoundError
from spotml.providers.gcp.helpers.gs_client import GSClient
from spotml.providers.gcp.resources.bucket import Bucket
from spotml.utils import random_string


class BucketManager(AbstractBucketManager):

    def __init__(self, project_name: str, region: str):
        super().__init__(project_name)

        self._gs = GSClient()
        self._region = region
        self._bucket_prefix = 'spotml-%s' % project_name.lower()

    def get_bucket(self) -> Bucket:
        buckets = self._gs.list_buckets()

        regex = re.compile('-'.join([self._bucket_prefix, '[a-z0-9]{12}', self._region]))
        buckets = [bucket for bucket in buckets if regex.match(bucket.name) is not None]

        if len(buckets) > 1:
            raise ValueError('Found several project buckets in the same region: %s.'
                             % ', '.join(bucket.name for bucket in buckets))

        if not len(buckets):
            raise BucketNotFoundError

        bucket = Bucket(buckets[0])

        return bucket

    def create_bucket(self) -> Bucket:
        bucket_name = '-'.join([self._bucket_prefix, random_string(12), self._region])
        bucket = self._gs.create_bucket(bucket_name, self._region)

        return Bucket(bucket)



import os
from schema import Schema, Optional, And, Regex, Or, Use
from spotml.config.validation import validate_config, get_instance_parameters_schema, has_prefix
from spotml.constants.constants import SPOT_STRATEGY_ON_DEMAND, SPOT_STRATEGY_SPOT


def validate_instance_parameters(params: dict):
    from spotml.providers.aws.config.ebs_volume import EbsVolume

    instance_parameters = {
        'region': And(str, Regex(r'^[a-z0-9-]+$')),
        Optional('availabilityZone', default=''): And(str, Regex(r'^[a-z0-9-]+$')),
        Optional('subnetId', default=''): And(str, Regex(r'^subnet-[a-z0-9]+$')),
        'instanceType': str,
        # Optional('spotInstance', default=False): bool,
        Optional('spotStrategy', default=SPOT_STRATEGY_ON_DEMAND): str,
        Optional('amiName', default=None): And(str, len, Regex(r'^[\w\(\)\[\]\s\.\/\'@-]{3,128}$')),
        Optional('amiId', default=None): And(str, len, Regex(r'^ami-[a-z0-9]+$')),
        Optional('rootVolumeSize', default=0): And(Or(int, str), Use(str),
                                                   Regex(r'^\d+$', error='Incorrect value for "rootVolumeSize".'),
                                                   Use(int),
                                                   And(lambda x: x > 0,
                                                       error='"rootVolumeSize" should be greater than 0 or should '
                                                             'not be specified.'),
                                                   ),
        Optional('ports', default=[]): [And(int, lambda x: 0 < x < 65536)],
        Optional('maxPrice', default=0): And(Or(float, int, str), Use(str),
                                             Regex(r'^\d+(\.\d{1,6})?$', error='Incorrect value for "maxPrice".'),
                                             Use(float),
                                             And(lambda x: x > 0, error='"maxPrice" should be greater than 0 or '
                                                                        'should  not be specified.'),
                                             ),
        Optional('managedPolicyArns', default=[]): [str],
        Optional('instanceProfileArn', default=None): str,
    }

    volumes_checks = [
        And(lambda x: len(x) < 12, error='Maximum 11 volumes are supported at the moment.'),
        And(lambda x: not has_prefix([(volume['parameters']['mountDir'] + '/') for volume in x
                                      if volume['parameters'].get('mountDir')]),
            error='Mount directories cannot be prefixes for each other.'),
    ]

    instance_checks = [
        And(lambda x: not (x['maxPrice'] and not x['spotStrategy'] == SPOT_STRATEGY_SPOT),
            error='"maxPrice" can be specified only for spot instances.'),
        And(lambda x: not (x['amiName'] and x['amiId']),
            error='"amiName" and "amiId" parameters cannot be used together.'),
    ]

    schema = get_instance_parameters_schema(instance_parameters, EbsVolume.TYPE_NAME, instance_checks, volumes_checks)

    return validate_config(schema, params)


def validate_ebs_volume_parameters(params: dict):
    from spotml.providers.aws.config.ebs_volume import EbsVolume

    old_deletion_policies_map = {
        'create_snapshot': EbsVolume.DP_CREATE_SNAPSHOT,
        'update_snapshot': EbsVolume.DP_UPDATE_SNAPSHOT,
        'retain': EbsVolume.DP_RETAIN,
        'delete': EbsVolume.DP_DELETE,
    }

    schema = Schema({
        Optional('volumeName', default=''): And(str, Regex(r'^[\w-]{1,255}$')),
        Optional('mountDir', default=''): And(
            str,
            And(os.path.isabs, error='Use absolute paths in the "mountDir" parameters'),
            Use(lambda x: x.rstrip('/'))
        ),
        Optional('size', default=0): And(int, lambda x: x > 0),
        # TODO: add the "iops" parameter to support the "io1" EBS volume type
        Optional('type', default='gp2'): lambda x: x in ['gp2', 'sc1', 'st1', 'standard'],
        Optional('deletionPolicy', default=EbsVolume.DP_RETAIN): And(
            str,
            lambda x: x in [EbsVolume.DP_CREATE_SNAPSHOT,
                            EbsVolume.DP_UPDATE_SNAPSHOT,
                            EbsVolume.DP_RETAIN,
                            EbsVolume.DP_DELETE] + list(old_deletion_policies_map.keys()),
            Use(lambda x: old_deletion_policies_map.get(x, x)),
            error='Incorrect value for "deletionPolicy".',
        ),
    })

    return validate_config(schema, params)



from argparse import ArgumentParser, Namespace

from spotml.commands.abstract_config_command import AbstractConfigCommand
from spotml.commands.writers.abstract_output_writrer import AbstractOutputWriter
from spotml.constants.run_status import STOPPED_DUE_TO_ERROR
from spotml.deployment.abstract_ssh_instance_manager import AbstractSshInstanceManager
from spotml.services.run_service import update_run_status


class HandleErroredRunCommand(AbstractConfigCommand):
    name = 'handle-errored-run'
    description = 'Run a custom script from the configuration file inside the container'

    def configure(self, parser: ArgumentParser):
        super().configure(parser)
        parser.add_argument('script_name', metavar='SCRIPT_NAME', type=str, help='Script name')
        parser.add_argument('-u', '--user', type=str, default=None,
                            help='Container username or UID (format: <name|uid>[:<group|gid>])')
        parser.add_argument('-s', '--session-name', type=str, default=None, help='tmux session name')
        parser.add_argument('-l', '--logging', action='store_true', help='Log the script outputs to a file')
        parser.add_argument('-p', '--parameter', metavar='PARAMETER=VALUE', action='append', type=str, default=[],
                            help='Set a value for the script parameter (format: PARAMETER=VALUE). This '
                                 'argument can be used multiple times to set several parameters. Parameters can be '
                                 'used in the script as Mustache variables (for example: {{PARAMETER}}).')
        parser.add_argument('--no-sync', action='store_true', help='Don\'t sync the project before running the script')

        # add the "double-dash" argument to the usage message
        parser.prog = 'spotml run'
        parser.usage = parser.format_usage()[7:-1] + ' [-- args...]\n'
        parser.epilog = 'The double dash (--) separates custom arguments that you can pass to the script ' \
                        'from the spotml arguments.'

    def _run(self, instance_manager: AbstractSshInstanceManager, args: Namespace, output: AbstractOutputWriter):
        if instance_manager.is_running():
            print("Errored instance is still running, stopping it now")
            instance_manager.stop(only_shutdown=False, output=output)
            update_run_status(args.run_id, STOPPED_DUE_TO_ERROR)
            return
        else:
            print("Errored instance is not running, Nothing to do.")
            update_run_status(args.run_id, STOPPED_DUE_TO_ERROR)



from argparse import ArgumentParser, Namespace

from spotml.commands.abstract_config_command import AbstractConfigCommand
from spotml.commands.writers.abstract_output_writrer import AbstractOutputWriter
from spotml.deployment.abstract_instance_manager import AbstractInstanceManager
from spotml.errors.nothing_to_do import NothingToDoError
from spotml.services.run_service import track_instance_start, track_instance_stop


class ManageCommand(AbstractConfigCommand):
    name = 'run'
    description = 'Run a custom script remotely managed by SpotML'

    def configure(self, parser: ArgumentParser):
        super().configure(parser)
        parser.add_argument('script_name', metavar='SCRIPT_NAME', type=str, help='Script name')
        parser.add_argument('-u', '--user', type=str, default=None,
                            help='Container username or UID (format: <name|uid>[:<group|gid>])')
        parser.add_argument('-s', '--session-name', type=str, default=None, help='tmux session name')
        parser.add_argument('-l', '--logging', action='store_true', help='Log the script outputs to a file')
        parser.add_argument('-p', '--parameter', metavar='PARAMETER=VALUE', action='append', type=str, default=[],
                            help='Set a value for the script parameter (format: PARAMETER=VALUE). This '
                                 'argument can be used multiple times to set several parameters. Parameters can be '
                                 'used in the script as Mustache variables (for example: {{PARAMETER}}).')
        parser.add_argument('--no-sync', action='store_true', help='Don\'t sync the project before running the script')

        # add the "double-dash" argument to the usage message
        parser.prog = 'spotml run'
        parser.usage = parser.format_usage()[7:-1] + ' [-- args...]\n'
        parser.epilog = 'The double dash (--) separates custom arguments that you can pass to the script ' \
                        'from the spotml arguments.'

    def _run(self, instance_manager: AbstractInstanceManager, args: Namespace, output: AbstractOutputWriter):
        # check that the script exists
        script_name = args.script_name
        if script_name == "stop":
            if instance_manager.is_running():
                track_instance_start(instance_manager.ssh_key_path)
                print(
                    'Note that you need to manually ssh and kill any running commands. Use the "spotml sh run" '
                    'command to connect to the running session.\n')
            else:
                track_instance_stop()
            return

        scripts = instance_manager.project_config.scripts
        if script_name not in scripts:
            raise ValueError('Script "%s" is not defined in the configuration file.' % script_name)

        instance_manager.check_dockerfile_exists()
        instance_manager.create_or_get_bucket(output, False)
        instance_manager.maybe_create_key()

        # sync the project with the instance
        if not args.no_sync:
            try:
                instance_manager.sync_to_bucket(output, False)
            except NothingToDoError:
                pass

        instance_manager.sendConfigCredentialFiles(args.script_name)
        output.write("Scheduled run on the cloud. Use 'spotml status' to check the status of the run. \n")



from abc import ABC
from spotml.deployment.abstract_cloud_instance.resources.abstract_bucket import AbstractBucket


class AbstractBucketManager(ABC):

    def __init__(self, project_name: str):
        self._project_name = project_name

    @property
    def project_name(self) -> str:
        return self._project_name

    def get_bucket(self) -> AbstractBucket:
        raise NotImplementedError

    def create_bucket(self) -> AbstractBucket:
        raise NotImplementedError



from spotml.providers.aws.resources.subnet import Subnet
from spotml.providers.aws.resources.vpc import Vpc


def get_vpc_id(ec2, subnet_id: str = None) -> str:
    """Returns VPC ID that should be used for deployment."""
    if subnet_id:
        vpc_id = Subnet.get_by_id(ec2, subnet_id).vpc_id
    else:
        default_vpc = Vpc.get_default_vpc(ec2)
        if not default_vpc:
            raise ValueError('Default VPC not found')

        vpc_id = default_vpc.vpc_id

    return vpc_id



from abc import ABC, abstractmethod
from spotml.config.abstract_instance_config import AbstractInstanceConfig


class AbstractContainerCommands(ABC):

    def __init__(self, instance_config: AbstractInstanceConfig):
        self._instance_config = instance_config

    @property
    def instance_config(self) -> AbstractInstanceConfig:
        return self._instance_config

    @abstractmethod
    def exec(self, command: str, interactive: bool = False, tty: bool = False, user: str = None,
             container_name: str = None, working_dir: str = None) -> str:
        raise NotImplementedError



from typing import List
from google.cloud import storage
from google.cloud.storage import Bucket


class GSClient(object):
    """Google Storage client."""

    def __init__(self):
        self._client = storage.Client()

    def list_buckets(self) -> List[Bucket]:
        res = list(self._client.list_buckets())
        return res

    def create_bucket(self, bucket_name: str, region: str) -> Bucket:
        bucket = Bucket(self._client, name=bucket_name)
        bucket.create(location=region)

        return bucket

    def create_dir(self, bucket_name: str, path: str):
        bucket = Bucket(self._client, name=bucket_name)
        blob = bucket.blob(path.rstrip('/') + '/')
        blob.upload_from_string('')



from spotml.config.validation import validate_config, get_instance_parameters_schema


def validate_instance_parameters(params: dict):
    from spotml.config.host_path_volume import HostPathVolume

    schema = get_instance_parameters_schema({}, HostPathVolume.TYPE_NAME)

    return validate_config(schema, params)



import re
import chevron
from spotml.deployment.utils.commands import get_bash_command


def parse_script_parameters(script_params: str):
    """Parses script parameters."""
    params = {}
    for param in script_params:
        match = re.match('(\\w+)=(.*)', param)
        if not match:
            raise ValueError('Invalid format for the script parameter: "%s" (the "PARAMETER=VALUE" format is expected).'
                             % param)

        param_name, param_value = match.groups()
        if param_name in params:
            raise ValueError('Parameter "%s" was defined twice.' % param_name)

        params[param_name] = param_value

    return params


def render_script(template: str, params: dict):
    """Renders a script template.

    It based on the Mustache templates, but only
    variables and delimiter changes are allowed.

    Raises an exception if one of the provided parameters doesn't
    exist in the template.
    """
    tokens = list(chevron.tokenizer.tokenize(template))
    template_keys = set()
    for tag, key in tokens:
        if tag not in ['literal', 'no escape', 'variable', 'set delimiter']:
            raise ValueError('Script templates support only variables and delimiter changes.')

        template_keys.add(key)

    # check that the script contains keys for all provided parameters
    for key in params:
        if key not in template_keys:
            raise ValueError('Parameter "%s" doesn\'t exist in the script.' % key)

    content = chevron.render(tokens, params)

    if content[:2] != '#!':
        content = ('#!%s\n\nset -xe\n\n' % get_bash_command()) + content

    return content



#!/usr/bin/env python3
import json
import subprocess

from datetime import datetime, timedelta
from dateutil.tz import tzutc
from dateutil.parser import parse as date_parser

intervals = (
    ('weeks', 604800),  # 60 * 60 * 24 * 7
    ('days', 86400),  # 60 * 60 * 24
    ('hours', 3600),  # 60 * 60
    ('minutes', 60),
    ('seconds', 1),
)

data = {}


def display_time(seconds, granularity=2):
    result = []

    for name, count in intervals:
        value = seconds // count
        if value:
            seconds -= value * count
            if value == 1:
                name = name.rstrip('s')
            result.append("{} {}".format(value, name))
    return ', '.join(result[:granularity])


def get_docker_container_ids():
    container_ids = []
    process = subprocess.Popen(["docker", "ps", "--format='{{json .}}'"], stdout=subprocess.PIPE,
                               stderr=subprocess.STDOUT)
    line_items = process.stdout.read().strip().split(b"\n")
    for element in line_items:
        element = element.decode().strip()[1:-1]
        if not element:
            continue
        metadata = json.loads(element)
        container_ids.append(metadata['ID'])
    return container_ids


def get_a_running_command(container_id: str):
    process = subprocess.Popen(["docker", "top", container_id], stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
    line_items = process.stdout.read().strip().split(b"\n")
    data["all_commands"] = []
    for element in line_items[1:]:
        command = element.split()[7:]
        command = b" ".join(command)
        command = command.decode()
        data["all_commands"].append(command)
        executable_command = command.split()[0].lower().split('/')[-1]
        bash_like_commands = ["command", "ps", "bash", "fish", "sh", "tsh", "tini", "zsh"]
        if executable_command not in bash_like_commands and 'jupyter-lab' not in command and 'ipykernel_launcher' not in command and "<defunct>" not in command:
            data['running_command'] = command
            return command
    return None


def get_least_idle_tty_time():
    current_time = datetime.now(tzutc())
    process = subprocess.Popen(["docker", "exec", container_id, "ls", "/dev/pts"], stdout=subprocess.PIPE,
                               stderr=subprocess.STDOUT)
    pseudo_terminals = process.stdout.read().decode('utf8').strip().split()
    data['pts_all'] = pseudo_terminals
    pseudo_terminals = [x for x in pseudo_terminals if x[0].isdigit()]
    least_idle_time_seconds = 60 * 60 * 24 * 365  # 1 year
    data['pts_timestamp_messages'] = []
    data['pts_last_seen_messages'] = []
    for pseudo_terminal in pseudo_terminals:
        process = subprocess.Popen(["docker", "exec", container_id, "stat", "/dev/pts/" + str(pseudo_terminal)],
                                   stdout=subprocess.PIPE,
                                   stderr=subprocess.STDOUT)
        line_items = process.stdout.read().strip().split(b"\n")

        for element in line_items:
            element = element.decode()
            accessors = element.split()
            if accessors[0] in ["Access:", "Modify:", "Change:"] and len(accessors) == 4:
                pts_timestamp_msg = "PTS %s was last %s at %s %s %s" % (
                    pseudo_terminal, accessors[0], accessors[1], accessors[2], accessors[3])
                data['pts_timestamp_messages'].append(pts_timestamp_msg)
                last_accessed_time = date_parser(accessors[1] + " " + accessors[2] + " " + accessors[3])
                idle_time_seconds = (current_time - last_accessed_time).total_seconds()
                data['pts_last_seen_messages'].append("TTY activity last seen %s ago" % display_time(idle_time_seconds))
                if idle_time_seconds < least_idle_time_seconds:
                    least_idle_time_seconds = idle_time_seconds
    return least_idle_time_seconds


docker_container_ids = get_docker_container_ids()
data["container_ids"] = docker_container_ids

running_command = False
for container_id in docker_container_ids:
    running_command = get_a_running_command(container_id)
    has_running_command = True if running_command else False
    idle_time_seconds = get_least_idle_tty_time()

data["running_command"] = running_command
data["idle_time_seconds"] = idle_time_seconds
data["idle_time"] = display_time(idle_time_seconds)

print(json.dumps(data))



from argparse import ArgumentParser, Namespace
from spotml.commands.abstract_config_command import AbstractConfigCommand
from spotml.commands.writers.abstract_output_writrer import AbstractOutputWriter
from spotml.deployment.utils.commands import get_bash_command, get_tmux_session_command
from spotml.errors.instance_not_running import InstanceNotRunningError
from spotml.deployment.abstract_instance_manager import AbstractInstanceManager


def get_session_name(args):
    session_name = 'spotml-sh-container'
    if args.session_name:
        session_name = args.session_name
    elif args.run:
        session_name = args.run
    return session_name


class ShCommand(AbstractConfigCommand):
    name = 'sh'
    description = 'Get a shell to the container or to the instance itself'

    def configure(self, parser: ArgumentParser):
        parser.add_argument('run', metavar='run', nargs='?', type=str, help='Ssh into run tmux session')
        super().configure(parser)
        parser.add_argument('-u', '--user', type=str, default=None,
                            help='Container username or UID (format: <name|uid>[:<group|gid>])')
        parser.add_argument('-H', '--host-os', action='store_true', help='Connect to the host OS instead of the Docker '
                                                                         'container')
        parser.add_argument('-s', '--session-name', type=str, default=None, help='tmux session name')
        parser.add_argument('-l', '--list-sessions', action='store_true', help='List all tmux sessions managed by the '
                                                                               'instance')

    def _run(self, instance_manager: AbstractInstanceManager, args: Namespace, output: AbstractOutputWriter):

        # check that the instance is started
        if not instance_manager.is_running():
            raise InstanceNotRunningError(instance_manager.instance_config.name)

        if args.list_sessions:
            if not instance_manager.use_tmux:
                raise ValueError('The "%s" provider doesn\'t support tmux.'
                                 % instance_manager.instance_config.provider_name)

            # a command to list existing tmux session on the host OS
            command = 'tmux ls; echo ""'
        else:
            if args.host_os:
                # get a command to open a login shell on the host OS
                session_name = args.session_name if args.session_name else 'spotml-sh-host'
                shell_command = '$SHELL'
                command = get_tmux_session_command(shell_command, session_name, keep_pane=False) \
                    if instance_manager.use_tmux else shell_command
            else:
                # get a command to run bash inside the docker container
                command = instance_manager.container_commands.exec(get_bash_command(), interactive=True, tty=True,
                                                                   user=args.user)

                # wrap the command with the tmux session
                if instance_manager.use_tmux:
                    session_name = get_session_name(args)

                    command = get_tmux_session_command(command, session_name, default_command=command, keep_pane=False)

        # execute command on the host OS
        instance_manager.exec(command)



import os
import subprocess
from spotml.configuration import get_spotml_keys_dir
from shutil import which
from spotml.providers.instance_manager_factory import PROVIDER_GCP


class SshKeyManager(object):

    def __init__(self, project_name: str, zone: str):
        self._key_name = 'spotml-key'
        self._keys_dir = get_spotml_keys_dir(PROVIDER_GCP)

    @property
    def private_key_file(self):
        return os.path.join(self._keys_dir, self._key_name)

    @property
    def public_key_file(self):
        return os.path.join(self._keys_dir, self._key_name + '.pub')

    def get_public_key_value(self):
        # generate a key if it doesn't exist
        if not os.path.isfile(self.private_key_file) or not os.path.isfile(self.public_key_file):
            self._generate_ssh_key()

        # read the public key value
        with open(self.public_key_file, 'r') as f:
            public_key_value = f.read().split()[1]

        return public_key_value

    def _generate_ssh_key(self):
        # delete the private key file if it already exists
        if os.path.isfile(self.private_key_file):
            os.unlink(self.private_key_file)

        # create a provider subdirectory
        if not os.path.isdir(self._keys_dir):
            os.makedirs(self._keys_dir, mode=0o755, exist_ok=True)

        # check that the "ssh-keygen" tool is installed
        ssh_keygen_cmd = 'ssh-keygen'
        if which(ssh_keygen_cmd) is None:
            raise ValueError('"ssh-keygen" command not found.')

        generate_key_cmd = [ssh_keygen_cmd, '-t', 'rsa', '-N', '', '-f', self.private_key_file, '-q']

        # generate a key pair
        res = subprocess.run(generate_key_cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        if res.returncode:
            raise subprocess.CalledProcessError(res.returncode, generate_key_cmd)



from argparse import ArgumentParser, Namespace
from spotml.commands.abstract_config_command import AbstractConfigCommand
from spotml.commands.writers.abstract_output_writrer import AbstractOutputWriter
from spotml.deployment.abstract_instance_manager import AbstractInstanceManager
from spotml.services.run_service import print_run_status


class StatusCommand(AbstractConfigCommand):
    name = 'status'
    description = 'Print information about the instance'

    def configure(self, parser: ArgumentParser):
        super().configure(parser)
        parser.add_argument('-l', '--logs', action='store_true', help='Print the latest run logs')

    def _run(self, instance_manager: AbstractInstanceManager, args: Namespace, output: AbstractOutputWriter):
        output.write("INSTANCE STATUS:")
        output.write(instance_manager.get_status_text())
        output.write("\nLAST RUN STATUS:")
        print_run_status(args.logs)



#!/usr/bin/env python

import os
import re
from setuptools import setup, find_packages

REQUIRED = []
with open('requirements.txt') as f:
    for line in f.readlines():
        REQUIRED.append(line.replace('\n', ''))
REQUIRES_PYTHON = '>=3.6.0'


def get_version():
    root_dir = os.path.abspath(os.path.dirname(__file__))
    with open(os.path.join(root_dir, 'spotml', '__init__.py')) as f:
        content = f.read()

    version_match = re.search(r'^__version__ = [\'"]([^\'"]*)[\'"]', content, re.M)
    if not version_match:
        raise RuntimeError('Unable to find version string.')

    return version_match.group(1)


def get_description():
    readme_path = os.path.abspath(os.path.join(os.path.dirname(__file__), 'Pypi.md'))
    with open(readme_path, encoding='utf-8') as f:
        description = f.read()

    return description


setup(name='spotml',
      version=get_version(),
      description='Automate ML training on spot instances easily.',
      url='https://spotml.io/',
      author='Vishnu',
      author_email='vishnu@spotml.io',
      long_description=get_description(),
      long_description_content_type='text/markdown',
      install_requires=REQUIRED,
      python_requires=REQUIRES_PYTHON,
      entry_points={
          'console_scripts': ['spotml=spotml.cli:main']
      },
      packages=find_packages(exclude=['tests*']),
      package_data={
          'spotml.deployment.container.docker.scripts': ['data/*', 'data/*/*'],
          'spotml.providers.aws.cfn_templates.instance': ['data/*', 'data/*/*'],
          'spotml.providers.aws.cfn_templates.instance_profile': ['data/*', 'data/*/*'],
          'spotml.providers.gcp.dm_templates.instance': ['data/*', 'data/*/*'],
      },
      tests_require=['moto'],
      test_suite='tests',
      classifiers=[
          'Development Status :: 4 - Beta',
          'Intended Audience :: Science/Research',
          'Intended Audience :: Developers',
          'Intended Audience :: System Administrators',
          'Natural Language :: English',
          'License :: OSI Approved :: Apache Software License',
          'Programming Language :: Python',
          'Programming Language :: Python :: 3',
          'Programming Language :: Python :: 3.6',
          'Programming Language :: Python :: 3.7',
      ])



import subprocess
from abc import ABC, abstractmethod
from spotml.commands.writers.abstract_output_writrer import AbstractOutputWriter
from spotml.config.abstract_instance_config import AbstractInstanceConfig
from spotml.config.project_config import ProjectConfig
from spotml.deployment.container.abstract_container_commands import AbstractContainerCommands


class AbstractInstanceManager(ABC):

    def __init__(self, project_config: ProjectConfig, instance_config: dict):
        self._project_config = project_config
        self._instance_config = self._get_instance_config(instance_config)

    @property
    def project_config(self) -> ProjectConfig:
        return self._project_config

    @property
    def instance_config(self) -> AbstractInstanceConfig:
        return self._instance_config

    @property
    def ssh_key_path(self):
        raise NotImplementedError

    @abstractmethod
    def _get_instance_config(self, instance_config: dict) -> AbstractInstanceConfig:
        """A factory method to create a provider's instance config."""
        raise NotImplementedError

    @property
    @abstractmethod
    def container_commands(self) -> AbstractContainerCommands:
        """A collection of commands to manage a container from the host OS."""
        raise NotImplementedError

    @abstractmethod
    def is_running(self) -> bool:
        """Checks if the instance is running."""
        raise NotImplementedError

    @abstractmethod
    def start(self, output: AbstractOutputWriter, dry_run=False):
        """Creates a stack with the instance."""
        raise NotImplementedError

    @abstractmethod
    def start_container(self, output: AbstractOutputWriter, dry_run=False):
        """Starts or restarts container on the host OS."""
        raise NotImplementedError

    @abstractmethod
    def stop(self, only_shutdown: bool, output: AbstractOutputWriter):
        """Deletes the stack."""
        raise NotImplementedError

    def exec(self, command: str, tty: bool = True) -> int:
        """Executes a command on the host OS."""
        result = subprocess.run(command, shell=True)
        return result.returncode

    def exec_capture_output(self, command: str) -> subprocess.CompletedProcess:
        """Executes a command on the host OS."""
        return subprocess.run(command, shell=True, stdout=subprocess.PIPE)

    @abstractmethod
    def clean(self, output: AbstractOutputWriter):
        """Deletes the stack."""
        raise NotImplementedError

    @abstractmethod
    def sync(self, output: AbstractOutputWriter, dry_run=False):
        """Synchronizes the project code with the instance."""
        raise NotImplementedError

    @abstractmethod
    def sync_to_bucket(self, output: AbstractOutputWriter, dry_run=False):
        """Synchronizes the project code with the instance."""
        raise NotImplementedError

    @abstractmethod
    def create_or_get_bucket(self, output: AbstractOutputWriter, dry_run: bool = False):
        """Synchronizes the project code with the instance."""
        raise NotImplementedError

    @abstractmethod
    def sync_to_instance(self, output: AbstractOutputWriter, dry_run=False):
        """Synchronizes the project code with the instance."""
        raise NotImplementedError

    @abstractmethod
    def download(self, download_filters: list, output: AbstractOutputWriter, dry_run=False):
        """Downloads files from the instance."""
        raise NotImplementedError

    @abstractmethod
    def get_status_text(self) -> str:
        """Returns information about the started instance.

        It will be shown to the user once the instance is started and by using the "status" command.
        """
        raise NotImplementedError

    @abstractmethod
    def sendConfigCredentialFiles(self) -> None:
        """Gets spotml config and credentials needed to run commands remotely."""
        raise NotImplementedError

    @abstractmethod
    def check_dockerfile_exists(self) -> None:
        raise NotImplementedError

    @abstractmethod
    def maybe_create_key(self) -> None:
        raise NotImplementedError

    @property
    def use_tmux(self) -> bool:
        """Use tmux when running a custom script or connecting to the instance."""
        return False



import json
import logging
from argparse import ArgumentParser, Namespace

from spotml.commands.abstract_config_command import AbstractConfigCommand
from spotml.commands.writers.abstract_output_writrer import AbstractOutputWriter
from spotml.constants.constants import INTERNAL_TAG
from spotml.constants.run_status import STOPPED_DUE_TO_IDLE_TIME, WAITING_FOR_SPOT, RUNNING, \
    STOPPED_UNKNOWN_REASON
from spotml.deployment.abstract_ssh_instance_manager import AbstractSshInstanceManager
from spotml.services.run_service import update_run_status, get_run


class HandleActiveRunCommand(AbstractConfigCommand):
    name = 'handle-active-run'
    description = 'Run a custom script from the configuration file inside the container'

    def configure(self, parser: ArgumentParser):
        super().configure(parser)
        parser.add_argument('script_name', metavar='SCRIPT_NAME', type=str, help='Script name')
        parser.add_argument('-u', '--user', type=str, default=None,
                            help='Container username or UID (format: <name|uid>[:<group|gid>])')
        parser.add_argument('-s', '--session-name', type=str, default=None, help='tmux session name')
        parser.add_argument('-i', '--max-idle-minutes', type=int,
                            help='Maximum Idle minutes to wait before turning off '
                                 'instance.')
        parser.add_argument('-l', '--logging', action='store_true', help='Log the script outputs to a file')
        parser.add_argument('-p', '--parameter', metavar='PARAMETER=VALUE', action='append', type=str, default=[],
                            help='Set a value for the script parameter (format: PARAMETER=VALUE). This '
                                 'argument can be used multiple times to set several parameters. Parameters can be '
                                 'used in the script as Mustache variables (for example: {{PARAMETER}}).')
        parser.add_argument('--no-sync', action='store_true', help='Don\'t sync the project before running the script')

        # add the "double-dash" argument to the usage message
        parser.prog = 'spotml run'
        parser.usage = parser.format_usage()[7:-1] + ' [-- args...]\n'
        parser.epilog = 'The double dash (--) separates custom arguments that you can pass to the script ' \
                        'from the spotml arguments.'

    def _run(self, instance_manager: AbstractSshInstanceManager, args: Namespace, output: AbstractOutputWriter):
        run = get_run(args.run_id)
        # check that the instance is started
        if not instance_manager.is_running():
            logging.info("Instance is already not running.")
            if run['status'] == RUNNING and run['command'] == "run":
                logging.info("It was probably killed by spot instance recall")
                update_run_status(args.run_id, WAITING_FOR_SPOT)
            if run['status'] == RUNNING and run['command'] == "start":
                logging.info("It was probably killed by spot instance recall")
                update_run_status(args.run_id, STOPPED_UNKNOWN_REASON)
            return

        result = instance_manager.exec_capture_output(
            'python3 /tmp/spotml/instance/scripts/startup/07_check_host_status.py')

        host_status = json.loads(result.stdout)
        print(f'{INTERNAL_TAG} {json.dumps(host_status)}')
        print(f"{INTERNAL_TAG} Idle Time Seconds: {host_status['idle_time_seconds']}")
        print(f"Detected Instance Idle Time: {host_status['idle_time']}")
        print(f"Detected Running Command: {host_status['running_command']}")
        print(f"Configured Max Idle minutes: {args.max_idle_minutes}")
        if not host_status['running_command'] and float(host_status['idle_time_seconds']) > args.max_idle_minutes * 60:
            print(f"Stopping instance due to idle time being longer than {args.max_idle_minutes} mins")
            instance_manager.stop(only_shutdown=False, output=output)
            update_run_status(args.run_id, STOPPED_DUE_TO_IDLE_TIME)



from abc import ABC, abstractmethod


class AbstractDataTransfer(ABC):

    def __init__(self, local_project_dir: str, host_project_dir: str, sync_filters: list, instance_name: str):
        self._instance_name = instance_name
        self._local_project_dir = local_project_dir
        self._host_project_dir = host_project_dir
        self._sync_filters = sync_filters

    @property
    def instance_name(self):
        return self._instance_name

    @property
    @abstractmethod
    def scheme_name(self) -> str:
        raise NotImplementedError

    def _get_bucket_project_path(self, bucket_name: str) -> str:
        """A bucket path where the project files are located."""
        return '%s://%s/project' % (self.scheme_name, bucket_name)

    def _get_bucket_downloads_path(self, bucket_name: str) -> str:
        """A bucket path where the downloaded files are located."""
        return '%s://%s/download/instance-%s' % (self.scheme_name, bucket_name, self.instance_name)

    @abstractmethod
    def upload_local_to_bucket(self, bucket_name: str, dry_run: bool = False):
        """Uploads files from local to the bucket."""
        raise NotImplementedError

    @abstractmethod
    def download_bucket_to_local(self, bucket_name: str, download_filters: list):
        """Downloads files from the bucket to local."""
        raise NotImplementedError

    @abstractmethod
    def get_download_bucket_to_instance_command(self, bucket_name: str, use_sudo: bool = False) -> str:
        """A remote command to download files from the bucket to the instance."""
        raise NotImplementedError

    @abstractmethod
    def get_upload_instance_to_bucket_command(self, bucket_name: str, download_filters: list, use_sudo: bool = False,
                                              dry_run: bool = False) -> str:
        """A remote command to upload files from the instance to the bucket."""
        raise NotImplementedError



import sys
from argparse import ArgumentParser, Namespace
from spotml.commands.abstract_config_command import AbstractConfigCommand
from spotml.commands.writers.abstract_output_writrer import AbstractOutputWriter
from spotml.deployment.utils.cli import shlex_join
from spotml.errors.instance_not_running import InstanceNotRunningError
from spotml.errors.nothing_to_do import NothingToDoError
from spotml.deployment.abstract_instance_manager import AbstractInstanceManager


class ExecCommand(AbstractConfigCommand):

    name = 'exec'
    description = 'Execute a command in the container'

    def configure(self, parser: ArgumentParser):
        super().configure(parser)
        parser.add_argument('-i', '--interactive', action='store_true', help='Pass STDIN to the container')
        parser.add_argument('-t', '--tty', action='store_true', help='Allocate a pseudo-TTY')
        parser.add_argument('-u', '--user', type=str, default=None,
                            help='Container username or UID (format: <name|uid>[:<group|gid>])')
        parser.add_argument('--no-sync', action='store_true', help='Don\'t sync the project before running the script')

        # add the "double-dash" argument to the usage message
        parser.prog = 'spotml exec'
        parser.usage = parser.format_usage()[7:-1] + ' -- COMMAND [args...]\n'
        parser.epilog = 'The double dash (--) separates the command that you want to execute inside the container ' \
                        'from the spotml arguments.'

    def _run(self, instance_manager: AbstractInstanceManager, args: Namespace, output: AbstractOutputWriter):
        # check that the command is provided
        if not args.custom_args:
            raise ValueError('Use the double-dash ("--") to split spotml arguments from the command that should be '
                             'executed inside the container.')

        # check that the instance is started
        if not instance_manager.is_running():
            raise InstanceNotRunningError(instance_manager.instance_config.name)

        # sync the project with the instance
        if not args.no_sync:
            try:
                instance_manager.sync(output)
            except NothingToDoError:
                pass

        # generate a "docker exec" command
        command = shlex_join(args.custom_args)
        command = instance_manager.container_commands.exec(command, interactive=args.interactive, tty=args.tty,
                                                           user=args.user)

        # execute the command on the host OS
        exit_code = instance_manager.exec(command, tty=args.tty)
        sys.exit(exit_code)



SCHEDULED = 'SCHEDULED'
RUNNING = 'RUNNING'
ERROR = 'ERROR'
LAUNCHING_INSTANCE = 'LAUNCHING_INSTANCE'
STOPPED_BY_USER = 'STOPPED_BY_USER'
STOPPED_DUE_TO_IDLE_TIME = 'STOPPED_DUE_TO_IDLE_TIME'
STOPPED_DUE_TO_ERROR = 'STOPPED_DUE_TO_ERROR'
WAITING_FOR_SPOT = 'WAITING_FOR_SPOT'
STOPPED_UNKNOWN_REASON = 'STOPPED_UNKNOWN_REASON'



from schema import And, Optional
from spotml.config.validation import validate_config, get_instance_parameters_schema


def validate_instance_parameters(params: dict):
    from spotml.config.host_path_volume import HostPathVolume

    instance_parameters = {
        'user': str,
        'host': str,
        Optional('port', default=22): And(int, lambda x: 0 < x < 65536),
        'keyPath': str,
    }

    schema = get_instance_parameters_schema(instance_parameters, HostPathVolume.TYPE_NAME)

    return validate_config(schema, params)



import os
from collections import namedtuple
import yaml
from spotml.config.project_config import ProjectConfig
from spotml.config.validation import DEFAULT_CONTAINER_NAME


DEFAULT_CONFIG_FILENAME = 'spotml.yaml'
OVERRIDE_CONFIG_FILENAME = 'spotml.override.yaml'


def load_config(config_path: str = None) -> ProjectConfig:
    # get project directory
    if not config_path:
        config_path = DEFAULT_CONFIG_FILENAME

    if os.path.isabs(config_path):
        config_abs_path = config_path
    else:
        config_abs_path = os.path.abspath(os.path.join(os.getcwd(), config_path))

    if not os.path.exists(config_abs_path):
        raise ValueError('Configuration file "%s" not found.' % config_path)

    # get the project directory
    project_dir = os.path.dirname(config_abs_path)

    # read the config
    config = _read_yaml(config_abs_path)

    # update the config if an override config exists
    if os.path.basename(config_abs_path) == DEFAULT_CONFIG_FILENAME:
        override_config_abs_path = os.path.join(project_dir, OVERRIDE_CONFIG_FILENAME)
        if os.path.isfile(override_config_abs_path):
            override_config = _read_yaml(override_config_abs_path)
            config = _merge_configs(config, override_config)

    # get project configuration
    project_config = ProjectConfig(config, project_dir)

    return project_config


def _read_yaml(file_path: str):
    """Returns content of the YAML file."""
    with open(file_path, 'r') as f:
        res = yaml.safe_load(f)

    return res


def _merge_configs(orig_config, override_config):
    """Merges original config with the override config."""

    MergeRule = namedtuple('MergeRule', ['key', 'merge_key', 'default_value', 'has_default_value'])

    merge_rules = [MergeRule(
        key='containers',
        merge_key='name',
        default_value=DEFAULT_CONTAINER_NAME,
        has_default_value=True,
    ), MergeRule(
        key='instances',
        merge_key='name',
        default_value=None,
        has_default_value=False,
    )]

    # validate and merge lists by keys
    for rule in merge_rules:
        if override_config and (rule.key in orig_config) and (rule.key in override_config):
            if not isinstance(orig_config[rule.key], list):
                raise ValueError('The "%s" key in the config must be a list.' % rule.key)

            if not isinstance(override_config[rule.key], list):
                raise ValueError('The "%s" key in the override config must be a list.' % rule.key)

            # convert lists to dictionaries
            dicts_to_merge = []
            for list_to_merge in [orig_config[rule.key], override_config[rule.key]]:
                dict_to_merge = {}
                for item in list_to_merge:
                    if not isinstance(item, dict):
                        raise ValueError('Each item of the "%s" list must be a dictionary.' % rule.key)

                    if rule.merge_key in item:
                        merge_value = item[rule.merge_key]
                    elif rule.has_default_value:
                        merge_value = rule.default_value
                    else:
                        raise ValueError('Each item of the "%s" list must contain the "%s" field.'
                                         % (rule.key, rule.merge_key))

                    if merge_value in dict_to_merge:
                        raise ValueError('Each item of the "%s" list must have a unique "%s" value.'
                                         % (rule.key, rule.merge_key))

                    dict_to_merge[merge_value] = item

                dicts_to_merge.append(dict_to_merge)

            # merge lists
            merged_dict = _update_dict(*dicts_to_merge)
            orig_config[rule.key] = [{**item, rule.merge_key: key} for key, item in merged_dict.items()]
            del override_config[rule.key]

    # merge the rest of the override config
    config = _update_dict(orig_config, override_config)

    return config


def _update_dict(d, u):
    if not isinstance(u, dict):
        return d

    if not isinstance(d, dict):
        return u

    for k, v in u.items():
        if isinstance(d, dict):
            if isinstance(v, dict):
                d[k] = _update_dict(d.get(k, {}), v)
            else:
                d[k] = u[k]
        else:
            d = {k: u[k]}

    return d



import os
from abc import ABC, abstractmethod
from collections import OrderedDict, namedtuple
from typing import List

from spotml.config.abstract_instance_volume import AbstractInstanceVolume
from spotml.config.container_config import ContainerConfig
from spotml.config.project_config import ProjectConfig
from spotml.config.tmp_dir_volume import TmpDirVolume
from spotml.config.validation import DEFAULT_CONTAINER_NAME, is_subdir
from spotml.deployment.abstract_cloud_instance.file_structure import CONTAINERS_TMP_DIR
from spotml.utils import filter_list

VolumeMount = namedtuple('VolumeMount', ['name', 'host_path', 'mount_path', 'mode', 'hidden'])


class AbstractInstanceConfig(ABC):

    def __init__(self, instance_config: dict, project_config: ProjectConfig):
        self._project_config = project_config

        # set instance parameters
        self._name = instance_config['name']
        self._provider_name = instance_config['provider']
        self._params = self._validate_instance_params(instance_config['parameters'])

        # get container config
        container_configs = filter_list(project_config.containers, 'name', self.container_name)
        if not container_configs:
            raise ValueError('Container configuration with the name "%s" not found.' % self.container_name)

        self._container_config = ContainerConfig(container_configs[0])

        # get volumes
        self._volumes = self._get_volumes()

        # get container volume mounts
        self._volume_mounts = self._get_volume_mounts(self._volumes)

        # get the host project directory
        self._host_project_dir = self._get_host_project_dir(self._volume_mounts)

    @abstractmethod
    def _validate_instance_params(self, params: dict) -> dict:
        """Validates instance parameters and fill missing ones with the default values."""
        raise NotImplementedError

    @abstractmethod
    def _get_instance_volumes(self) -> List[AbstractInstanceVolume]:
        """Returns specific to the provider volumes that should be mounted on the host OS."""
        raise NotImplementedError

    @property
    def project_config(self) -> ProjectConfig:
        return self._project_config

    @property
    def container_config(self) -> ContainerConfig:
        return self._container_config

    @property
    @abstractmethod
    def user(self) -> str:
        raise NotImplementedError

    @property
    def name(self) -> str:
        """Name of the instance."""
        return self._name

    @property
    def provider_name(self):
        """Provider name."""
        return self._provider_name

    @property
    def container_name(self) -> str:
        return self._params['containerName'] if self._params['containerName'] else DEFAULT_CONTAINER_NAME

    @property
    def full_container_name(self) -> str:
        """A container name that is used in the "docker run" command."""
        return ('spotml-%s-%s-%s' % (self.project_config.project_name, self.name, self.container_name)).lower()

    @property
    def docker_data_root(self) -> str:
        """Data root directory for Docker daemon."""
        return self._params['dockerDataRoot']

    @property
    def local_ssh_port(self) -> int:
        """Local SSH port to connect to the instance (in case of a tunnel)."""
        return self._params['localSshPort']

    @property
    def commands(self) -> str:
        """Commands that should be run once an instance is started."""
        return self._params['commands']

    @property
    def host_project_dir(self):
        """Project directory on the host OS."""
        return self._host_project_dir

    @property
    def volumes(self) -> List[AbstractInstanceVolume]:
        return self._volumes

    @property
    def volume_mounts(self) -> List[VolumeMount]:
        return self._volume_mounts

    @property
    def dockerfile_path(self):
        """Dockerfile path on the host OS."""
        dockerfile_path = self.container_config.file
        if dockerfile_path:
            dockerfile_path = self.host_project_dir + '/' + dockerfile_path

        return dockerfile_path

    @property
    def docker_context_path(self):
        """Docker build's context path on the host OS."""
        dockerfile_path = self.dockerfile_path
        if not dockerfile_path:
            return ''

        return os.path.dirname(dockerfile_path)

    @property
    def host_container_dir(self):
        """A temporary directory on the host OS that contains container-related files and directories."""
        return '%s/%s' % (CONTAINERS_TMP_DIR, self.full_container_name)

    @property
    def host_logs_dir(self):
        """A directory mainly for the "spotml run" command logs."""
        return self.host_container_dir + '/logs'

    @property
    def host_volumes_dir(self):
        """A directory with temporary volumes. If there is a Volume Mount in the configuration file
        that doesn't have a corresponding instance volume, a temporary directory will be created
        and attached to the container.
        """
        return self.host_container_dir + '/volumes'

    def _get_volumes(self) -> List[AbstractInstanceVolume]:
        """Returns volumes that should be mounted on the host OS."""
        volumes = self._get_instance_volumes()

        # create temporary volumes for the volume mounts that don't have corresponding
        # volumes in the instance configuration
        instance_volume_names = set(volume.name for volume in volumes)
        for container_volume in self.container_config.volume_mounts:
            if container_volume['name'] not in instance_volume_names:
                volumes.append(TmpDirVolume(volume_config={
                    'name': container_volume['name'],
                    'parameters': {'path': '%s/%s' % (self.host_volumes_dir, container_volume['name'])}
                }))

        return volumes

    def _get_volume_mounts(self, volumes: List[AbstractInstanceVolume]) \
            -> List[VolumeMount]:
        """Returns container volume mounts and a path to the project directory on the host OS."""
        # get mount directories for the volumes
        host_paths = OrderedDict([(volume.name, volume.host_path) for volume in volumes])

        # get container volumes mapping
        volume_mounts = []
        for container_volume in self.container_config.volume_mounts:
            volume_mounts.append(VolumeMount(
                name=container_volume['name'],
                host_path=host_paths[container_volume['name']],
                mount_path=container_volume['mountPath'],
                mode='rw',
                hidden=False,
            ))

        return volume_mounts

    def _get_host_project_dir(self, volume_mounts: List[VolumeMount]) -> str:
        """Returns the host project directory."""
        host_project_dir = None
        for volume_mount in sorted(volume_mounts, key=lambda x: len(x.mount_path), reverse=True):
            if is_subdir(self.container_config.project_dir, volume_mount.mount_path):
                # the project directory is a subdirectory of a Volume Mount directory
                project_subdir = os.path.relpath(self.container_config.project_dir, volume_mount.mount_path)
                host_project_dir = os.path.normpath(volume_mount.host_path + '/' + project_subdir)
                break

        # this should not be the case as the volume mount for the project directory should be added automatically
        # if it doesn't exist in the configuration
        assert host_project_dir is not None, 'A volume mount that contains the project directory not found.'

        return host_project_dir



from datetime import datetime
from spotml.deployment.abstract_cloud_instance.resources.abstract_instance import AbstractInstance
from spotml.providers.gcp.helpers.ce_client import CEClient


class Instance(AbstractInstance):

    def __init__(self, ce: CEClient, data: dict):
        """
        Args:
            data (dict): Example:
               {'canIpForward': False,
                'cpuPlatform': 'Intel Haswell',
                'creationTimestamp': '2019-04-20T16:21:49.536-07:00',
                'deletionProtection': False,
                'description': '',
                'disks': [{'autoDelete': True,
                           'boot': True,
                           'deviceName': 'instance-1',
                           'guestOsFeatures': [{'type': 'VIRTIO_SCSI_MULTIQUEUE'}],
                           'index': 0,
                           'interface': 'SCSI',
                           'kind': 'compute#attachedDisk',
                           'licenses': ['https://www.googleapis.com/compute/v1/projects/debian-cloud/global/licenses/debian-9-stretch'],
                           'mode': 'READ_WRITE',
                           'source': 'https://www.googleapis.com/compute/v1/projects/spotml-221422/zones/us-east1-b/disks/instance-1',
                           'type': 'PERSISTENT'}],
                'id': '928537266896639843',
                'kind': 'compute#instance',
                'labelFingerprint': '42WmSpB8rSM=',
                'machineType': 'https://www.googleapis.com/compute/v1/projects/spotml-221422/zones/us-east1-b/machineTypes/n1-standard-1',
                'metadata': {'fingerprint': 'IoRxXrApBlw=',
                             'kind': 'compute#metadata'},
                'name': 'instance-1',
                'networkInterfaces': [{'accessConfigs': [{'kind': 'compute#accessConfig',
                                                          'name': 'External NAT',
                                                          'natIP': '34.73.140.188',
                                                          'networkTier': 'PREMIUM',
                                                          'type': 'ONE_TO_ONE_NAT'}],
                                       'fingerprint': 'COAWpxIgZx0=',
                                       'kind': 'compute#networkInterface',
                                       'name': 'nic0',
                                       'network': 'https://www.googleapis.com/compute/v1/projects/spotml-221422/global/networks/default',
                                       'networkIP': '10.142.0.2',
                                       'subnetwork': 'https://www.googleapis.com/compute/v1/projects/spotml-221422/regions/us-east1/subnetworks/default'}],
                'scheduling': {'automaticRestart': False,
                               'onHostMaintenance': 'TERMINATE',
                               'preemptible': True},
                'selfLink': 'https://www.googleapis.com/compute/v1/projects/spotml-221422/zones/us-east1-b/instances/instance-1',
                'serviceAccounts': [{'email': '293101887402-compute@developer.gserviceaccount.com',
                                     'scopes': ['https://www.googleapis.com/auth/devstorage.read_only',
                                                'https://www.googleapis.com/auth/logging.write',
                                                'https://www.googleapis.com/auth/monitoring.write',
                                                'https://www.googleapis.com/auth/servicecontrol',
                                                'https://www.googleapis.com/auth/service.management.readonly',
                                                'https://www.googleapis.com/auth/trace.append']}],
                'startRestricted': False,
                'status': 'RUNNING',
                'tags': {'fingerprint': '42WmSpB8rSM='},
                'zone': 'https://www.googleapis.com/compute/v1/projects/spotml-221422/zones/us-east1-b'}
        """
        self._ce = ce
        self._data = data

    @staticmethod
    def get_by_name(ce: CEClient, machine_name: str):
        """Returns an instance by its stack name."""
        res = ce.list_instances(machine_name)
        if not res:
            return None

        return Instance(ce, res[0])

    @property
    def name(self) -> str:
        return self._data['name']

    @property
    def is_running(self) -> bool:
        return self.status == 'RUNNING'

    @property
    def is_stopped(self) -> bool:
        # see Instance Life Cycle: https://cloud.google.com/compute/docs/instances/instance-life-cycle
        return self.status == 'TERMINATED'

    @property
    def public_ip_address(self) -> str:
        return self._data['networkInterfaces'][0]['accessConfigs'][0].get('natIP')

    @property
    def status(self) -> str:
        return self._data['status']

    @property
    def machine_type(self) -> str:
        return self._data['machineType'].split('/')[-1]

    @property
    def zone(self) -> str:
        return self._data['zone'].split('/')[-1]

    @property
    def creation_timestamp(self) -> datetime:
        # fix the format: '2019-04-20T16:21:49.536-07:00' -> '2019-04-20T16:21:49-0700'
        time_str = self._data['creationTimestamp'][:-10] + \
                   self._data['creationTimestamp'][-6:-3] + \
                   self._data['creationTimestamp'][-2:]
        return datetime.strptime(time_str, '%Y-%m-%dT%H:%M:%S%z')

    @property
    def is_preemtible(self) -> bool:
        return self._data['scheduling']['preemptible']

    def terminate(self, wait: bool = True):
        self._ce.delete_instance(self.name, wait)

    def stop(self, wait: bool = True):
        self._ce.stop_instance(self.name, wait)



from argparse import Namespace, ArgumentParser
from spotml.commands.abstract_config_command import AbstractConfigCommand
from spotml.commands.writers.abstract_output_writrer import AbstractOutputWriter
from spotml.errors.instance_not_running import InstanceNotRunningError
from spotml.errors.nothing_to_do import NothingToDoError
from spotml.deployment.abstract_instance_manager import AbstractInstanceManager


class DownloadCommand(AbstractConfigCommand):

    name = 'download'
    description = 'Download files from the running instance'

    def configure(self, parser: ArgumentParser):
        super().configure(parser)
        parser.add_argument('-i', '--include', metavar='PATTERN', action='append', type=str, required=True,
                            help='Download all files that matches the specified pattern (see Include Filters '
                                 'for the "aws s3 sync" command). Paths must be relative to your project directory, '
                                 'they cannot be absolute.')
        parser.add_argument('--dry-run', action='store_true', help='Show files to be downloaded')

    def _run(self, instance_manager: AbstractInstanceManager, args: Namespace, output: AbstractOutputWriter):
        # check that the instance is started
        if not instance_manager.is_running():
            raise InstanceNotRunningError(instance_manager.instance_config.name)

        filters = [
            {'exclude': ['*']},
            {'include': args.include}
        ]

        dry_run = args.dry_run
        with output.prefix('[dry-run] ' if dry_run else ''):
            try:
                instance_manager.download(filters, output, dry_run)
            except NothingToDoError as e:
                output.write(str(e))
                return

        output.write('Done')



import argparse
from typing import List, Type
import pkg_resources
from spotml.commands.abstract_command import AbstractCommand
from spotml.commands.aws import AwsCommand
from spotml.commands.download import DownloadCommand
from spotml.commands.exec import ExecCommand
from spotml.commands.handle_error_runs import HandleErroredRunCommand
from spotml.commands.run import ManageCommand
from spotml.commands.handle_scheduled_run import HandleScheduledRunCommand
from spotml.commands.run_locally import RunLocallyCommand
from spotml.commands.sh import ShCommand
from spotml.commands.start import StartCommand
from spotml.commands.status import StatusCommand
from spotml.commands.stop import StopCommand
from spotml.commands.handle_active_run import HandleActiveRunCommand
from spotml.commands.sync import SyncCommand


def get_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser()
    parser.add_argument('-V', '--version', action='store_true', help='Display the version of the spotml')

    command_classes = [
       StartCommand,
       StopCommand,
       StatusCommand,
       ShCommand,
       ManageCommand,
       RunLocallyCommand,
       HandleScheduledRunCommand,
       HandleActiveRunCommand,
       HandleErroredRunCommand,
       ExecCommand,
       SyncCommand,
       DownloadCommand,
       AwsCommand,
    ] + _get_custom_commands()

    # add commands to the parser
    add_subparsers(parser, command_classes)

    return parser


def add_subparsers(parser: argparse.ArgumentParser, command_classes: List[Type[AbstractCommand]]):
    """Adds commands to the parser."""
    subparsers = parser.add_subparsers()
    for command_class in command_classes:
        command = command_class()
        subparser = subparsers.add_parser(command.name, help=command.description, description=command.description)
        subparser.set_defaults(command=command, parser=subparser)
        command.configure(subparser)


def _get_custom_commands() -> List[Type[AbstractCommand]]:
    """Returns custom commands that integrated through entry points."""
    return [entry_point.load() for entry_point in pkg_resources.iter_entry_points('spotml.commands')]
